import json
import re
import zipfile
from pathlib import Path
from unittest import mock

import pytest
import responses

from cumulusci.core.config import TaskConfig
from cumulusci.core.config.marketing_cloud_service_config import (
    MarketingCloudServiceConfig,
)
from cumulusci.core.exceptions import DeploymentException
from cumulusci.tasks.marketing_cloud.deploy import (
    MCPM_BASE_ENDPOINT,
    MCPM_JOB_ID_HEADER,
    UNKNOWN_STATUS_MESSAGE,
    MarketingCloudDeployTask,
)
from cumulusci.tasks.marketing_cloud.mc_constants import MC_API_VERSION
from cumulusci.utils import temporary_dir

TEST_TSSD = "asdf-qwerty"
STACK_KEY = "S4"
TEST_ZIPFILE = "test-mc-pkg.zip"
PAYLOAD_FILE = "expected-payload.json"
JOB_NAME = "foo_job"

# This JSON does not necessarily include the complete data provided by Marketing Cloud
# It is intended to exercise functionality in the class under test.
# The response is designed to go with the content of `expected-payload.json`
VALIDATION_RESPONSE_FILE = "validation-response.json"


@pytest.fixture
def task(project_config):
    test_zip_file = Path(__file__).parent.absolute() / TEST_ZIPFILE
    task = MarketingCloudDeployTask(
        project_config,
        TaskConfig(
            {
                "options": {
                    "package_zip_file": test_zip_file.resolve(),
                    "custom_inputs": "companyName:Acme",
                    "name": JOB_NAME,
                }
            }
        ),
    )
    task.mc_config = MarketingCloudServiceConfig(
        {
            "rest_instance_url": f"https://{TEST_TSSD}.rest.marketingcloudapis.com/",
            "access_token": "foo",
        },
        "mc",
        None,
    )
    task.poll_interval_s = 0  # Do not poll during testing
    return task


@pytest.fixture
def validation_response():
    with open(Path(__file__).parent.absolute() / VALIDATION_RESPONSE_FILE) as f:
        return json.load(f)


@pytest.fixture
def expected_payload():
    expected_payload_file = Path(__file__).parent.absolute() / PAYLOAD_FILE
    with open(expected_payload_file, "r") as f:
        return json.load(f)


@pytest.fixture
def task_without_custom_inputs(project_config):
    test_zip_file = Path(__file__).parent.absolute() / TEST_ZIPFILE
    task = MarketingCloudDeployTask(
        project_config,
        TaskConfig(
            {
                "options": {
                    "package_zip_file": test_zip_file.resolve(),
                    "custom_inputs": None,
                    "name": JOB_NAME,
                }
            }
        ),
    )
    task.mc_config = MarketingCloudServiceConfig(
        {
            "rest_instance_url": f"https://{TEST_TSSD}.rest.marketingcloudapis.com/",
            "access_token": "foo",
        },
        "mc",
        None,
    )
    return task


@pytest.fixture
def mocked_responses():
    with responses.RequestsMock() as rsps:
        rsps.add(
            "GET",
            f"https://{TEST_TSSD}.auth.marketingcloudapis.com/{MC_API_VERSION}/userinfo",
            json={"organization": {"stack_key": STACK_KEY}},
        )
        yield rsps


@pytest.fixture
def mocked_validation_responses(mocked_responses):
    mocked_responses.add(
        "POST",
        f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/validate",
        json={"id": "0be865fe-efb2-479d-99c2-c1b608155369"},
        match=[responses.matchers.header_matcher({MCPM_JOB_ID_HEADER: JOB_NAME})],
    )
    mocked_responses.add(
        "GET",
        f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/validate/0be865fe-efb2-479d-99c2-c1b608155369",
        match=[responses.matchers.header_matcher({MCPM_JOB_ID_HEADER: JOB_NAME})],
        json={"status": "DONE"},
    )

    yield mocked_responses


@pytest.fixture
def mocked_polling_validation_responses(mocked_responses, validation_response):
    mocked_responses.add(
        "POST",
        f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/validate",
        match=[responses.matchers.header_matcher({MCPM_JOB_ID_HEADER: JOB_NAME})],
        json={"id": "0be865fe-efb2-479d-99c2-c1b608155369"},
    )
    mocked_responses.add(
        "GET",
        f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/validate/0be865fe-efb2-479d-99c2-c1b608155369",
        match=[responses.matchers.header_matcher({MCPM_JOB_ID_HEADER: JOB_NAME})],
        json={"status": "NOT_FOUND"},
    )
    mocked_responses.add(
        "GET",
        f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/validate/0be865fe-efb2-479d-99c2-c1b608155369",
        match=[responses.matchers.header_matcher({MCPM_JOB_ID_HEADER: JOB_NAME})],
        json={"status": "IN_PROGRESS"},
    )
    mocked_responses.add(
        "GET",
        f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/validate/0be865fe-efb2-479d-99c2-c1b608155369",
        match=[responses.matchers.header_matcher({MCPM_JOB_ID_HEADER: JOB_NAME})],
        json={"status": "NOT_FOUND"},
    )
    mocked_responses.add(
        "GET",
        f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/validate/0be865fe-efb2-479d-99c2-c1b608155369",
        match=[responses.matchers.header_matcher({MCPM_JOB_ID_HEADER: JOB_NAME})],
        json=validation_response,
    )

    yield mocked_responses


@pytest.fixture
def mocked_polling_validation_responses_all_not_found(mocked_responses):
    mocked_responses.add(
        "POST",
        f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/validate",
        match=[responses.matchers.header_matcher({MCPM_JOB_ID_HEADER: JOB_NAME})],
        json={"id": "0be865fe-efb2-479d-99c2-c1b608155369"},
    )
    for _ in range(10):
        mocked_responses.add(
            "GET",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/validate/0be865fe-efb2-479d-99c2-c1b608155369",
            match=[responses.matchers.header_matcher({MCPM_JOB_ID_HEADER: JOB_NAME})],
            json={"status": "NOT_FOUND"},
        )

    yield mocked_responses


class TestMarketingCloudDeployTask:
    def test_run_task__deploy_succeeds_with_custom_inputs(
        self, task, mocked_validation_responses
    ):
        mocked_validation_responses.add(
            "POST",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments",
            json={"id": "JOBID", "status": "IN_PROGRESS"},
        )
        mocked_validation_responses.add(
            "GET",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments/JOBID",
            json={"status": "DONE", "entities": {}},
        )

        task.logger = mock.Mock()
        task._run_task()
        task.logger.info.assert_called_with(
            "Deployment (JOBID) completed successfully."
        )
        assert task.logger.error.call_count == 0
        assert task.logger.warn.call_count == 0

    def test_run_task__deploy_succeeds_without_custom_inputs(
        self, task_without_custom_inputs, mocked_validation_responses
    ):
        mocked_validation_responses.add(
            "POST",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments",
            json={"id": "JOBID", "status": "IN_PROGRESS"},
        )
        mocked_validation_responses.add(
            "GET",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments/JOBID",
            json={"status": "DONE", "entities": {}},
        )
        task = task_without_custom_inputs
        task.logger = mock.Mock()
        task._run_task()
        task.logger.info.assert_called_with(
            "Deployment (JOBID) completed successfully."
        )
        assert task.logger.error.call_count == 0
        assert task.logger.warn.call_count == 0

    def test_run_task__deploy_fails(self, task, mocked_validation_responses):
        mocked_validation_responses.add(
            "POST",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments",
            json={"id": "JOBID", "status": "IN_PROGRESS"},
        )
        mocked_validation_responses.add(
            "GET",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments/JOBID",
            json={
                "status": "DONE",
                "entities": {
                    "assets": {
                        0: {"status": "FAILED", "issues": ["A problem occurred"]},
                        1: {"status": "SKIPPED", "issues": ["A problem occurred"]},
                    },
                    "categories": {},
                },
            },
        )
        task.logger = mock.Mock()
        with pytest.raises(DeploymentException):
            task._run_task()

        assert task.logger.error.call_count == 2
        assert (
            task.logger.error.call_args_list[0][0][0]
            == "Failed to deploy assets/0. Status: FAILED. Issues: ['A problem occurred']"
        )
        assert (
            task.logger.error.call_args_list[1][0][0]
            == "Failed to deploy assets/1. Status: SKIPPED. Issues: ['A problem occurred']"
        )

    def test_run_task__FATAL_ERROR_result(self, task, mocked_validation_responses):
        mocked_validation_responses.add(
            "POST",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments",
            json={"id": "JOBID", "status": "FATAL_ERROR"},
        )
        mocked_validation_responses.add(
            "GET",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments/JOBID",
            json={
                "status": "FATAL_ERROR",
                "entities": {},
            },
        )
        task.logger = mock.Mock()
        with pytest.raises(DeploymentException):
            task._run_task()

    def test_run_task__unknown_deploy_status(
        self, task, mocked_validation_responses, caplog
    ):
        unknown_status = "FOOBAR"
        mocked_validation_responses.add(
            "POST",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments",
            json={"id": "JOBID"},
        )
        mocked_validation_responses.add(
            "GET",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments/JOBID",
            json={
                "status": unknown_status,
                "entities": {},
            },
        )
        caplog.clear()
        with pytest.raises(DeploymentException):
            task._run_task()

        logged_messages = [log.message for log in caplog.records]
        assert UNKNOWN_STATUS_MESSAGE.format(unknown_status) in logged_messages

    # NOTE: the x-mcpm-job-id header is validated implicitly by all use
    # of Responses mocks.

    def test_validate_package__validation_accepts_intermittent_not_found(
        self, task, mocked_polling_validation_responses
    ):
        mocked_polling_validation_responses.add(
            "POST",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments",
            json={"id": "JOBID", "status": "IN_PROGRESS"},
        )
        mocked_polling_validation_responses.add(
            "GET",
            f"{MCPM_BASE_ENDPOINT.format(STACK_KEY)}/deployments/JOBID",
            json={"status": "DONE", "entities": {}},
        )
        task._run_task()
        assert task.validation_response["status"] == "DONE"

    def test_validate_package__validation_fails_consistent_not_found(
        self, task, mocked_polling_validation_responses_all_not_found
    ):
        with pytest.raises(DeploymentException):
            task._run_task()

    def test_update_payload_entities_with_actions(
        self, task, expected_payload, validation_response
    ):
        task.validation_response = validation_response
        validated_payload = task._update_payload_entities_with_actions(expected_payload)

        assert validated_payload["entities"]["assets"]["0"]["action"] == {
            "type": "create",
            "available": True,
            "issues": [],
        }

    def test_action_for_entity(self, task, validation_response):
        task.validation_response = validation_response

        # We should expect to get back the first action with available == True
        assert task.action_for_entity("assets", "0") == {
            "type": "create",
            "available": True,
            "issues": [],
        }
        assert task.action_for_entity("foo", "0") is None

    def test_zipfile_not_valid(self, task):
        task.options["package_zip_file"] = "not-a-valid-file.zip"
        task.logger = mock.Mock()
        task._run_task()

        task.logger.error.assert_called_once_with(
            "Package zip file not valid: not-a-valid-file.zip"
        )

    def test_construct_payload(self, task, expected_payload):
        task.options["name"] = "cci-deploy"
        pkg_zip_file = Path(task.options["package_zip_file"])
        with temporary_dir() as temp_dir:
            with zipfile.ZipFile(pkg_zip_file) as zf:
                zf.extractall(temp_dir)
                actual_payload = task._construct_payload(Path(temp_dir))

        assert expected_payload == actual_payload

    def test_construct_payload__file_not_found(self, task):
        """Ensure we state clearly state where expect files to be"""
        task.options["name"] = "cci-deploy"
        pkg_zip_file = Path(task.options["package_zip_file"])
        with temporary_dir() as temp_dir:
            with zipfile.ZipFile(pkg_zip_file) as zf:
                zf.extractall(temp_dir)

                expected_payload_file = Path(temp_dir + "/info.json")
                assert expected_payload_file.is_file()
                Path.unlink(expected_payload_file)

                with pytest.raises(DeploymentException):
                    task._construct_payload(Path(temp_dir))

    def test_add_custom_inputs_to_payload__deployment_exception(self, task):
        custom_inputs = {"foo": "bar"}
        payload = {"input": [{"key": "baz"}]}
        error_message = re.escape("Custom input of key foo not found in package.")
        with pytest.raises(DeploymentException, match=error_message):
            task._add_custom_inputs_to_payload(custom_inputs, payload)

    def test_add_custom_inputs_to_payload(self, task):
        custom_inputs = {"companyName": "Acme"}
        payload = {"input": [{"key": "companyName"}]}
        payload = task._add_custom_inputs_to_payload(custom_inputs, payload)
        assert payload == {"input": [{"key": "companyName", "value": "Acme"}]}



from cumulusci.core.tasks import BaseTask
from cumulusci.tasks.util import Sleep


class ExampleTask(BaseTask):
    def _run_task(self):
        self.logger.info("Called _run_task")


class StaticPreflightTask(BaseTask):

    task_options = {
        "task_name": {
            "description": "Task that this preflight is for",
            "required": True,
        },
        "status_code": {
            "description": 'Status code to return, default to "ok".',
            "required": False,
        },
        "msg": {"description": "Message to be included", "required": False},
    }

    def _run_task(self):
        self.return_values["status_code"] = self.options.get("status_code", "ok")
        self.return_values["task_name"] = self.options.get("task_name")
        self.return_values["msg"] = self.options.get("msg", "")


class StaticSleep(Sleep):
    task_options = dict(
        **Sleep.task_options,
        **{
            "task_name": {
                "description": "Task that this preflight is for",
                "required": True,
            },
        }
    )

    def _run_task(self):
        super()._run_task()
        self.return_values["task_name"] = self.options.get("task_name")



import glob
import os
import shutil
import time

from defusedxml.minidom import parse

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.tasks import BaseTask
from cumulusci.utils import download_extract_zip, find_replace, find_replace_regex
from cumulusci.utils.options import (
    CCIOptions,
    DirectoryPath,
    Field,
    ListOfStringsOption,
)


class DownloadZip(BaseTask):
    name = "Download"

    class Options(CCIOptions):
        url: str = Field(..., description="The url of the zip file to download")
        dir: str = Field(
            ..., description="The directory where the zip should be extracted"
        )
        subfolder: str = Field(
            None,
            description=(
                "The subfolder of the target zip to extract. Defaults to"
                + " extracting the root of the zip file to the destination."
            ),
        )

    parsed_options: Options

    def _run_task(self):
        if not os.path.exists(self.parsed_options.dir):
            os.makedirs(self.parsed_options.dir)

        download_extract_zip(
            self.parsed_options.url,
            self.parsed_options.dir,
            self.parsed_options.subfolder,
        )


class ListMetadataTypes(BaseTask):
    name = "ListMetadataTypes"

    class Options(CCIOptions):
        package_xml: str = Field(
            None,
            description=(
                "The project package.xml file."
                + " Defaults to <project_root>/src/package.xml"
            ),
        )

    def _init_options(self, kwargs):
        super(ListMetadataTypes, self)._init_options(kwargs)
        if not self.parsed_options.get("package_xml"):
            self.parsed_options["package_xml"] = os.path.join(
                self.project_config.repo_root, "src", "package.xml"
            )

    def _run_task(self):
        dom = parse(self.parsed_options["package_xml"])
        package = dom.getElementsByTagName("Package")[0]
        types = package.getElementsByTagName("types")
        type_list = []
        for t in types:
            name = t.getElementsByTagName("name")[0]
            metadata_type = name.firstChild.nodeValue
            type_list.append(metadata_type)
        self.logger.info(
            "Metadata types found in %s:\r\n%s",
            self.parsed_options["package_xml"],
            "\r\n".join(type_list),
        )


class Sleep(BaseTask):
    name = "Sleep"

    class Options(CCIOptions):
        seconds: int = Field(..., description="The number of seconds to sleep")

    parsed_options: Options

    def _run_task(self):
        self.logger.info("Sleeping for {} seconds".format(self.parsed_options.seconds))
        time.sleep(float(self.parsed_options.seconds))
        self.logger.info("Done")


class Delete(BaseTask):
    name = "Delete"

    class Options(CCIOptions):
        path: ListOfStringsOption = Field(
            default=...,
            description=(
                "The path to delete. "
                "If path is a directory, recursively deletes the directory: BE CAREFUL!!! "
                "If path is a list, all paths will be deleted"
            ),
        )
        chdir: DirectoryPath = Field(
            default=None,
            description=(
                "Change directories before deleting path(s). "
                "This is useful if you have a common list of relative paths to "
                "delete that you want to call against different directories."
            ),
        )

    def _run_task(self):
        chdir = self.parsed_options.chdir
        cwd = os.getcwd()
        if chdir:
            self.logger.info("Changing directory to {}".format(chdir))
            os.chdir(chdir)

        path = self.parsed_options.path
        if not isinstance(path, list):
            path = [path]
        for path_item in path:
            matches = glob.glob(path_item)
            if matches:
                for match in matches:
                    self._delete(match)
            else:
                self.logger.info("{} does not exist, skipping delete".format(path))

        if chdir:
            os.chdir(cwd)

    def _delete(self, path):
        if os.path.isdir(path):
            self.logger.info("Recursively deleting directory {}".format(path))
            shutil.rmtree(path)
        else:
            self.logger.info("Deleting file {}".format(path))
            os.remove(path)


class FindReplaceOptions(CCIOptions):
    find: str = Field(..., description="The string to search for")
    replace: str = Field(
        "",
        description="The string to replace matches with. Defaults to an empty string",
    )
    path: DirectoryPath = Field(..., description="The path to recursively search")
    file_pattern: ListOfStringsOption = Field(
        "*",
        description="A UNIX like filename pattern used for matching filenames, or a list of them. See python fnmatch docs for syntax. If passed via command line, use a comma separated string. Defaults to *",
    )
    env_replace: bool = Field(
        False,
        description="If True, treat the value of the replace option as the name of an environment variable, and use the value of that variable as the replacement string. Defaults to False",
    )


class FindReplace(BaseTask):
    class Options(FindReplaceOptions):
        max: int = Field(
            None,
            description="The max number of matches to replace.  Defaults to replacing all matches.",
        )

    parsed_options: Options

    def _run_task(self):
        kwargs = {}
        if self.parsed_options.max:
            kwargs["max"] = self.parsed_options.max

        if self.parsed_options["env_replace"]:
            if self.parsed_options["replace"] in os.environ.keys():
                self.parsed_options["replace"] = os.environ[
                    self.parsed_options["replace"]
                ]
            else:
                raise TaskOptionsError(
                    f"The environment variable {self.parsed_options['replace']} was not found. Ensure that this value is populated or set env_replace to False."
                )

        for file_pattern in self.parsed_options["file_pattern"]:
            find_replace(
                find=self.parsed_options.find,
                replace=self.parsed_options.replace,
                directory=self.parsed_options.path,
                filePattern=file_pattern,
                logger=self.logger,
                **kwargs,
            )


class FindReplaceRegex(FindReplace):
    class Options(FindReplaceOptions):
        pass

    def _run_task(self):
        find_replace_regex(
            find=self.parsed_options.find,
            replace=self.parsed_options.replace,
            directory=self.parsed_options.path,
            filePattern=self.parsed_options.file_pattern,
            logger=self.logger,
        )


class CopyFile(BaseTask):
    task_options = {
        "src": {"description": "The path to the source file to copy", "required": True},
        "dest": {
            "description": "The destination path where the src file should be copied",
            "required": True,
        },
    }

    def _run_task(self):
        self.logger.info("Copying file {src} to {dest}".format(**self.options))
        shutil.copyfile(src=self.options["src"], dst=self.options["dest"])


class LogLine(BaseTask):
    task_options = {
        "level": {"description": "The logger level to use", "required": True},
        "line": {"description": "A formatstring like line to log", "required": True},
        "format_vars": {"description": "A Dict of format vars", "required": False},
    }

    def _init_options(self, kwargs):
        super(LogLine, self)._init_options(kwargs)
        if "format_vars" not in self.options:
            self.options["format_vars"] = {}

    def _run_task(self):
        log = getattr(self.logger, self.options["level"])
        log(self.options["line"].format(**self.options["format_vars"]))


class PassOptionAsResult(BaseTask):
    task_options = {
        "result": {"description": "The result for the task", "required": True}
    }

    def _run_task(self):
        return self.options["result"]


class PassOptionAsReturnValue(BaseTask):
    task_options = {
        "key": {"required": True, "description": "The return value key to use."},
        "value": {"required": True, "description": "The value to set."},
    }

    def _run_task(self):
        self.return_values[self.options["key"]] = self.options["value"]



# -*- coding: utf-8 -*-
import os
import re
from pprint import pformat
from unittest import mock

import pytest
from rich.emoji import Emoji
from rich.table import Column

from cumulusci.cli.ui import (
    CHECKMARK,
    CROSSMARK,
    CliTable,
    SimpleSalesforceUIHelpers,
    _summarize,
)


@pytest.fixture
def sample_data():
    return {
        "service_list": [
            ["Name", "Description", "Configured"],
            ["saucelabs", "Configure connection for saucelabs tasks.", False],
            ["sentry", "Configure connection to sentry.io for error tracking", False],
        ],
        "org_list": [
            [
                "Org",
                "Default",
                "Days",
                "Expired",
                "Config",
                Column("Domain", overflow="crop"),
            ],
            ["beta", None, 1, None, "beta", ""],
            ["dev", True, "1/7", False, "dev", "test-h8znvutwnctb@example.com"],
            ["feature", None, 1, None, "feature", ""],
        ],
        "task_list_util": [
            ["Task", "Description"],
            ["command", "Run an arbitrary command"],
            ["log", "Log a line at the info level."],
            ["util_sleep", "Sleeps for N seconds"],
        ],
    }


def test_table_dim_rows(sample_data):
    data = sample_data["service_list"]
    instance = CliTable(data, dim_rows=[1])
    assert instance.table.rows[0].style.dim


def test_table_stringify_booleans(sample_data):
    data = sample_data["service_list"]
    data[1][2] = True
    CliTable.PICTOGRAM_TRUE = CHECKMARK
    CliTable.PICTOGRAM_FALSE = CROSSMARK
    instance = CliTable(data)
    table = str(instance)
    if os.name == "posix":
        assert Emoji.replace(":heavy_check_mark:") in table
        assert Emoji.replace(":cross_mark:") in table
    else:
        assert "+" in table
        assert "-" in table


def test_table_stringify_misc(sample_data):
    data = sample_data["org_list"]
    instance = CliTable(data)
    table = str(instance)
    assert "7" in table
    assert "None" not in table


def test_columnify_headers():
    data = [["y", Column("n", overflow="crop"), 43], [1, 2, 3]]
    table = CliTable(data)
    str(table)
    formatted_table = str(table)
    assert "  y   n   43  " in formatted_table


def test_rich_truncation(sample_data):
    count = 6
    data = [["column"], ["x" * 16 * count]]
    table = CliTable(data, title="testerino", width=20)
    formatted_table = str(table)
    rxp = re.compile("x{16}")
    assert len(rxp.findall(formatted_table)) == count


PRETEND_DESCRIBE_DATA = {
    "name": "Blah",
    "updateable": False,
    "fields": [
        {"name": "Foo", "referenceTo": ["Bar"], "picklistValues": []},
        {
            "name": "Bar",
            "referenceTo": [],
            "picklistValues": [{"value": "a"}, {"value": "b"}],
        },
        {"name": "Baz", "type": "string", "referenceTo": [], "picklistValues": []},
    ],
}

SUMMARIZED_DATA = {"Foo": ["Bar"], "Bar": ["a", "b"], "Baz": "string"}


def test_summarize():
    fields = PRETEND_DESCRIBE_DATA["fields"]
    assert _summarize(fields[0]) == ("Foo", ["Bar"])
    assert _summarize(fields[1]) == ("Bar", ["a", "b"])
    assert _summarize(fields[2]) == ("Baz", "string")


def test_pretty_describe():
    sf = mock.MagicMock()
    sf.Blah.describe.return_value = PRETEND_DESCRIBE_DATA
    rc = SimpleSalesforceUIHelpers(sf).describe("Blah", detailed=False, format="obj")
    assert sf.Blah.describe.mock_calls
    assert rc == SUMMARIZED_DATA, rc


def test_pretty_describe_detailed():
    sf = mock.MagicMock()
    sf.Blah.describe.return_value = PRETEND_DESCRIBE_DATA
    rc = SimpleSalesforceUIHelpers(sf).describe("Blah", detailed=True, format="obj")
    assert sf.Blah.describe.mock_calls
    assert rc == PRETEND_DESCRIBE_DATA, rc


def test_pretty_describe_pprint(capsys):
    sf = mock.MagicMock()
    sf.Blah.describe.return_value = PRETEND_DESCRIBE_DATA
    SimpleSalesforceUIHelpers(sf).describe("Blah", detailed=False, format="pprint")
    out = capsys.readouterr().out
    assert sf.Blah.describe.mock_calls
    assert out.strip() == pformat(SUMMARIZED_DATA), (out, pformat(SUMMARIZED_DATA))


def test_pretty_describe_format_error():
    sf = mock.MagicMock()
    sf.Blah.describe.return_value = PRETEND_DESCRIBE_DATA
    with pytest.raises(TypeError):
        SimpleSalesforceUIHelpers(sf).describe("Blah", detailed=False, format="xyzzy")


def _pretend_soql_result(*args):
    return {
        "totalSize": 1,
        "done": True,
        "records": [{"attributes": {"type": "AggregateResult"}, "expr0": 20}],
    }


def _pretend_soql_results(num_records):
    rc = {"totalSize": 1, "done": True, "records": []}
    for i in range(0, num_records):
        rc["records"].append(
            {"attributes": {"type": "AggregateResult"}, "Id": "00GUSWLMA"}
        )
    return rc


def test_pretty_soql_query_simple_count():
    sf = mock.MagicMock()
    sf.query_all.return_value = _pretend_soql_result()
    rc = SimpleSalesforceUIHelpers(sf).query(
        "select Count(Id) from Account",
        include_deleted=False,
        format="obj",
        max_rows=100,
    )
    assert sf.query_all.mock_calls == [
        mock.call("select Count(Id) from Account", include_deleted=False)
    ], sf.query_all.mock_calls
    assert rc == [{"expr0": 20}]

    sf = mock.MagicMock()
    sf.query_all.return_value = _pretend_soql_result()

    rc = SimpleSalesforceUIHelpers(sf).query(
        "select Count(Id) from Account",
        include_deleted=True,
        format="obj",
        max_rows=100,
    )
    assert sf.query_all.mock_calls == [
        mock.call("select Count(Id) from Account", include_deleted=True)
    ], sf.query_all.mock_calls
    assert rc == [{"expr0": 20}]


def test_pretty_soql_query_simple_truncation():
    sf = mock.MagicMock()
    sf.query_all.return_value = _pretend_soql_results(150)
    rc = SimpleSalesforceUIHelpers(sf).query(
        "select Id from Account", include_deleted=False, format="obj", max_rows=100
    )
    assert sf.query_all.mock_calls == [
        mock.call("select Id from Account", include_deleted=False)
    ], sf.query_all.mock_calls
    assert len(rc) == 101, len(rc)
    assert rc[-1] == "... truncated 50 rows"


def test_pretty_soql_query_table(capsys):
    sf = mock.MagicMock()
    sf.query_all.return_value = _pretend_soql_result()

    SimpleSalesforceUIHelpers(sf).query(
        "select Count(Id) from Account",
        include_deleted=False,
        format="table",
        max_rows=100,
    )
    assert sf.query_all.mock_calls == [
        mock.call("select Count(Id) from Account", include_deleted=False)
    ], sf.query_all.mock_calls
    out = capsys.readouterr().out

    assert "expr0" in out
    assert "20" in out
    assert "help" in out


def test_pretty_soql_query_table_truncation(capsys):
    sf = mock.MagicMock()
    sf.query_all.return_value = _pretend_soql_results(150)

    SimpleSalesforceUIHelpers(sf).query(
        "select Id from Account",
        include_deleted=False,
        format="table",
        max_rows=100,
    )
    assert sf.query_all.mock_calls == [
        mock.call("select Id from Account", include_deleted=False)
    ], sf.query_all.mock_calls
    out = capsys.readouterr().out
    assert out.count("00GUSWLMA") == 100


def test_pretty_table_empty(capsys):
    sf = mock.MagicMock()
    sf.query_all.return_value = _pretend_soql_results(0)

    SimpleSalesforceUIHelpers(sf).query(
        "select Id from Account", include_deleted=False, format="table", max_rows=100
    )
    assert sf.query_all.mock_calls == [
        mock.call("select Id from Account", include_deleted=False)
    ], sf.query_all.mock_calls
    out = capsys.readouterr().out
    assert out.count("00GUSWLMA") == 0
    assert "No results" in out


def test_pretty_soql_query_pprint(capsys):
    sf = mock.MagicMock()
    sf.query_all.return_value = _pretend_soql_results(150)

    SimpleSalesforceUIHelpers(sf).query(
        "select Id from Account", include_deleted=False, format="pprint", max_rows=100
    )
    assert sf.query_all.mock_calls == [
        mock.call("select Id from Account", include_deleted=False)
    ], sf.query_all.mock_calls
    out = capsys.readouterr().out
    assert out.count("00GUSWLMA") == 100


def test_pretty_soql_query_json():
    sf = mock.MagicMock()
    sf.query_all.return_value = _pretend_soql_results(150)

    rc = SimpleSalesforceUIHelpers(sf).query(
        "select Id from Account", include_deleted=False, format="json", max_rows=100
    )
    assert sf.query_all.mock_calls == [
        mock.call("select Id from Account", include_deleted=False)
    ], sf.query_all.mock_calls

    assert rc.count("00GUSWLMA") == 100


def test_pretty_soql_query_errors():
    sf = mock.MagicMock()
    sf.query_all.return_value = _pretend_soql_results(150)

    with pytest.raises(TypeError):
        SimpleSalesforceUIHelpers(sf).query(
            "select Id from Account",
            include_deleted=False,
            format="punchcard",
            max_rows=100,
        )


@mock.patch("cumulusci.cli.ui.SimpleSalesforceUIHelpers.query")
@mock.patch("cumulusci.cli.ui.SimpleSalesforceUIHelpers.describe")
def test_repl_helpers(pretty_soql_query, pretty_describe):
    sf = mock.MagicMock()
    helpers = SimpleSalesforceUIHelpers(sf)
    assert helpers.describe("Account", format="obj")
    assert helpers.query("select Id from Account", format="obj")



from pathlib import Path

import yaml

from cumulusci.tests.pytest_plugins.pytest_sf_vcr_serializer import (
    CompressionVCRSerializer,
    RequestResponseReplacement,
    _compress_in_place,
)
from cumulusci.tests.pytest_plugins.vcr_string_compressor import (
    StringToTemplateCompressor,
)


class TestVCRSerializer:
    def test_compress_inclusions(self):
        serialized_request = "{'method': 'GET', 'uri': 'https://orgname.my.salesforce.com/services/data/vxx.0/sobjects/Organization/00D0xORGID00000000', 'body': None, 'headers': {'Request-Headers': ['Elided']}}"
        saved_responses = {
            serialized_request: RequestResponseReplacement(
                request=...,
                response=...,
                replacement_file="GET_sobjects_Organization.yaml",
            )
        }
        d = {
            "interactions": [
                {
                    "request": {
                        "method": "GET",
                        "uri": "https://orgname.my.salesforce.com/services/data/vxx.0/sobjects/Organization/00D0xORGID00000000",
                        "body": None,
                        "headers": {"Request-Headers": ["Elided"]},
                    },
                    "response": {"Blah"},
                }
            ],
            "version": 42,
        }
        x = _compress_in_place(d, saved_responses, StringToTemplateCompressor({}))
        assert x == {
            "interactions": [{"include_file": "GET_sobjects_Organization.yaml"}],
            "version": 42,
        }

    def test_compress_duplications(self):
        d = {
            "interactions": [
                {
                    "request": {
                        "method": "GET",
                        "uri": "https://orgname.my.salesforce.com/services/data/vxx.0/sobjects/Organization/00D0xORGID00000000",
                        "body": None,
                        "headers": {"Request-Headers": ["Elided"]},
                    },
                    "response": {"Blah": "Blah"},
                },
                {
                    "request": {
                        "method": "GET",
                        "uri": "https://orgname.my.salesforce.com/services/data/vxx.0/sobjects/Organization/00D0xORGID00000000",
                        "body": None,
                        "headers": {"Request-Headers": ["Elided"]},
                    },
                    "response": {"Blah": "Blah"},
                },
            ],
            "version": 42,
        }

        x = _compress_in_place(d, {}, StringToTemplateCompressor({}))
        assert x is d
        assert x["interactions"][0] is x["interactions"][1]

        print(yaml.dump(x))
        assert yaml.dump(x).strip() == EXPECTED.strip()

    def test_compress_file(self, cumulusci_test_repo_root):
        directory = cumulusci_test_repo_root / "cumulusci/tests/shared_cassettes"
        serializer = CompressionVCRSerializer(directory)
        file_to_compress = Path("cumulusci/tests/uncompressed_cassette.yaml")
        with file_to_compress.open() as f:
            data = yaml.safe_load(f)

        as_string = serializer.serialize(data)
        assert as_string.strip() == EXPECTED2.strip()


EXPECTED = """
interactions:
- &id001
  request:
    body: null
    headers:
      Request-Headers:
      - Elided
    method: GET
    uri: https://orgname.my.salesforce.com/services/data/vxx.0/sobjects/Organization/00D0xORGID00000000
  response:
    Blah: Blah
- *id001
version: 42""".strip()

EXPECTED2 = """
version: 1
interactions:
- &id001
  include_file: GET_sobjects_Global_describe.yaml
- include_file: GET_sobjects_Account_describe.yaml
- *id001
- request:
    method: POST
    uri: https://orgname.my.salesforce.com/services/async/vxx.0/job/75023000003Rs2qAAC
    body: <jobInfo xmlns="http://www.force.com/2009/06/asyncapi/dataload"><state>Closed</state></jobInfo>
    headers:
      Request-Headers:
      - Elided
  response:
    status:
      code: 200
      message: OK
    headers:
      Content-Type:
      - application/xml
      Others: Elided
    body:
      string:
        include_template: jobInfo_insert_xml.tpl
        vars:
          id: 75023000003Rs2qAAC
          creator: 00523000002KbrQAAS
          cdate: '2022-03-30T07:57:00.000Z'
          smts: '2022-03-30T07:57:00.000Z'
          state: Closed
          qdbatches: '1'"""



import json

from cumulusci.cli.ui import CliTable
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.core.utils import process_list_arg
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class AssignPermissionSets(BaseSalesforceApiTask):
    task_docs = """
Assigns Permission Sets whose Names are in ``api_names`` to either the default org user or the user whose Alias is ``user_alias``. This task skips assigning Permission Sets that are already assigned.
    """

    task_options = {
        "api_names": {
            "description": "API Names of desired Permission Sets, separated by commas.",
            "required": True,
        },
        "user_alias": {
            "description": "Target user aliases, separated by commas. Defaults to the current running user."
        },
    }

    permission_name = "PermissionSet"
    permission_name_field = "Name"
    permission_label = "Permission Set"
    assignment_name = "PermissionSetAssignment"
    assignment_lookup = "PermissionSetId"
    assignment_child_relationship = "PermissionSetAssignments"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        self.options["api_names"] = process_list_arg(self.options["api_names"])
        self.options["user_alias"] = process_list_arg(
            self.options.get("user_alias") or []
        )

    def _run_task(self):
        users = self._query_existing_assignments()
        users_assigned_perms = {
            user["Id"]: self._get_assigned_perms(user) for user in users
        }
        perms_by_id = self._get_perm_ids()

        records_to_insert = []
        for user_id, assigned_perms in users_assigned_perms.items():
            records_to_insert.extend(
                self._get_assignments(user_id, assigned_perms, perms_by_id)
            )

        self._insert_assignments(records_to_insert)

    def _query_existing_assignments(self):
        if not self.options["user_alias"]:
            query = (
                f"SELECT Id,(SELECT {self.assignment_lookup} FROM {self.assignment_child_relationship}) "
                "FROM User "
                f"WHERE Username = '{self.org_config.username}'"
            )
        else:
            aliases = "','".join(self.options["user_alias"])
            query = (
                f"SELECT Id,(SELECT {self.assignment_lookup} FROM {self.assignment_child_relationship}) "
                "FROM User "
                f"""WHERE Alias IN ('{aliases}')"""
            )

        result = self.sf.query(query)
        if result["totalSize"] == 0:
            raise CumulusCIException(
                "No Users were found matching the specified aliases."
            )
        return result["records"]

    def _get_assigned_perms(self, user):
        assigned_perms = {}
        # PermissionSetLicenseAssignments actually returns None if there are no assignments instead of an empty list of records.  Wow.
        if user[self.assignment_child_relationship]:
            assigned_perms = {
                r[self.assignment_lookup]
                for r in user[self.assignment_child_relationship]["records"]
            }
        return assigned_perms

    def _get_perm_ids(self):
        api_names = "', '".join(self.options["api_names"])
        perms = self.sf.query(
            f"SELECT Id,{self.permission_name_field} FROM {self.permission_name} WHERE {self.permission_name_field} IN ('{api_names}')"
        )
        perms_by_ids = {
            p["Id"]: p[self.permission_name_field] for p in perms["records"]
        }

        missing_perms = [
            api_name
            for api_name in self.options["api_names"]
            if api_name not in perms_by_ids.values()
        ]
        if missing_perms:
            raise CumulusCIException(
                f"The following {self.permission_label}s were not found: {', '.join(missing_perms)}."
            )
        return perms_by_ids

    def _get_assignments(self, user_id, assigned_perms, perms_by_id):
        assignments = []
        for perm, perm_name in perms_by_id.items():
            if perm not in assigned_perms:
                self.logger.info(
                    f'Assigning {self.permission_label} "{perm_name}" to {user_id}.'
                )
                assignment = {
                    "attributes": {"type": self.assignment_name},
                    "AssigneeId": user_id,
                    self.assignment_lookup: perm,
                }
                assignments.append(assignment)
            else:
                self.logger.warning(
                    f'{self.permission_label} "{perm_name}" is already assigned to {user_id}.'
                )
        return assignments

    def _insert_assignments(self, records_to_insert):
        result_list = []
        for i in range(0, len(records_to_insert), 200):
            request_body = json.dumps(
                {"allOrNone": False, "records": records_to_insert[i : i + 200]}
            )
            result = self.sf.restful(
                "composite/sobjects", method="POST", data=request_body
            )
            result_list.extend(result)
        self._process_composite_results(result_list)

    def _process_composite_results(self, api_results):
        results_table_data = [["Success", "ID", "Message"]]
        for result in api_results:
            result_row = [result["success"], result.get("id", "-")]
            if not result["success"] and result["errors"]:
                result_row.append(result["errors"][0]["message"])
            else:
                result_row.append("-")
            results_table_data.append(result_row)

        table = CliTable(
            results_table_data,
            title="Results",
        )
        table.echo()

        if not all([result["success"] for result in api_results]):
            raise CumulusCIException(
                f"Not all {self.assignment_child_relationship} were saved."
            )


class AssignPermissionSetLicenses(AssignPermissionSets):
    task_docs = """
Assigns Permission Set Licenses whose Developer Names or PermissionSetLicenseKey are in ``api_names`` to either the default org user or the user whose Alias is ``user_alias``. This task skips assigning Permission Set Licenses that are already assigned.

Permission Set Licenses are usually associated with a Permission Set, and assigning the Permission Set usually assigns the associated Permission Set License automatically.  However, in non-namespaced developer scratch orgs, assigning the associated Permission Set may not automatically assign the Permission Set License, and this task will ensure the Permission Set Licenses are assigned.
    """

    task_options = {
        "api_names": {
            "description": "API Developer Names of desired Permission Set Licenses, separated by commas.",
            "required": True,
        },
        "user_alias": {
            "description": "Alias of target user (if not the current running user, the default)."
        },
    }

    permission_name = "PermissionSetLicense"
    permission_name_field = ["DeveloperName", "PermissionSetLicenseKey"]
    permission_label = "Permission Set License"
    assignment_name = "PermissionSetLicenseAssign"
    assignment_lookup = "PermissionSetLicenseId"
    assignment_child_relationship = "PermissionSetLicenseAssignments"

    def _get_perm_ids(self):
        perms_by_ids = {}
        api_names = "', '".join(self.options["api_names"])
        perms = self.sf.query(
            f"SELECT Id,{self.permission_name_field[0]},{self.permission_name_field[1]} FROM {self.permission_name} WHERE {self.permission_name_field[0]} IN ('{api_names}') OR {self.permission_name_field[1]} IN ('{api_names}')"
        )
        for p in perms["records"]:
            if p[self.permission_name_field[0]] in self.options["api_names"]:
                perms_by_ids[p["Id"]] = p[self.permission_name_field[0]]
            else:
                perms_by_ids[p["Id"]] = p[self.permission_name_field[1]]

        missing_perms = [
            api_name
            for api_name in self.options["api_names"]
            if api_name not in perms_by_ids.values()
        ]
        if missing_perms:
            raise CumulusCIException(
                f"The following {self.permission_label}s were not found: {', '.join(missing_perms)}."
            )
        return perms_by_ids


class AssignPermissionSetGroups(AssignPermissionSets):
    task_docs = """
Assigns Permission Set Groups whose Developer Names are in ``api_names`` to either the default org user or the user whose Alias is ``user_alias``. This task skips assigning Permission Set Groups that are already assigned.
    """

    task_options = {
        "api_names": {
            "description": "API Developer Names of desired Permission Set Groups, separated by commas.",
            "required": True,
        },
        "user_alias": {
            "description": "Alias of target user (if not the current running user, the default)."
        },
    }

    permission_name = "PermissionSetGroup"
    permission_name_field = "DeveloperName"
    permission_label = "Permission Set Group"
    assignment_name = "PermissionSetAssignment"
    assignment_lookup = "PermissionSetGroupId"
    assignment_child_relationship = "PermissionSetAssignments"



import logging
from datetime import date
from io import StringIO
from pathlib import Path
from unittest import mock

import pytest
import responses

from cumulusci.core.exceptions import BulkDataException, YAMLParseException
from cumulusci.tasks.bulkdata.mapping_parser import (
    CaseInsensitiveDict,
    MappingLookup,
    MappingStep,
    ValidationError,
    _infer_and_validate_lookups,
    parse_from_yaml,
    validate_and_inject_mapping,
)
from cumulusci.tasks.bulkdata.step import DataApi, DataOperationType
from cumulusci.tests.util import DummyOrgConfig, mock_describe_calls


class TestMappingParser:
    def test_simple_parse(self):
        base_path = Path(__file__).parent / "mapping_v2.yml"
        assert parse_from_yaml(base_path)

    def test_after(self):
        base_path = Path(__file__).parent / "mapping_after.yml"
        result = parse_from_yaml(base_path)

        step = result["Insert Accounts"]
        lookups = step["lookups"]
        assert lookups
        assert "after" in lookups["ParentId"]
        after_list = {
            lookup["after"] for lookup in lookups.values() if "after" in lookup
        }
        assert after_list

    def test_deprecation(self, caplog):
        base_path = Path(__file__).parent / "mapping_v2.yml"
        caplog.set_level(logging.WARNING)

        parse_from_yaml(base_path)
        assert "record_type" in caplog.text

    def test_deprecation_override(self, caplog):
        base_path = Path(__file__).parent / "mapping_v2.yml"
        caplog.set_level(logging.WARNING)
        with mock.patch(
            "cumulusci.tasks.bulkdata.mapping_parser.SHOULD_REPORT_RECORD_TYPE_DEPRECATION",
            False,
        ):
            mapping = parse_from_yaml(base_path)
            assert "record_type" not in caplog.text
            assert mapping["Insert Households"]["record_type"] == "HH_Account"

    def test_bad_mapping_syntax(self):
        base_path = Path(__file__).parent / "mapping_v2.yml"
        with open(base_path, "r") as f:
            data = f.read().replace(":", ": abcd")
            with pytest.raises(
                YAMLParseException, match="An error occurred parsing yaml file .*"
            ):
                parse_from_yaml(StringIO(data))

    def test_bad_mapping_grammar(self):
        base_path = Path(__file__).parent / "mapping_v2.yml"
        with open(base_path, "r") as f:
            data = f.read().replace("record_type", "xyzzy")
            with pytest.raises(ValidationError):
                parse_from_yaml(StringIO(data))

    def test_bad_mapping_id_mode(self):
        base_path = Path(__file__).parent / "mapping_v2.yml"
        with open(base_path, "r") as f:
            data = f.read().replace("Name: name", "Id: sf_id")
            with pytest.raises(ValidationError):
                parse_from_yaml(StringIO(data))

    def test_bad_mapping_oid_as_pk(self):
        base_path = Path(__file__).parent / "mapping_v1.yml"
        with open(base_path, "r") as f:
            data = f.read().replace("api: bulk", "oid_as_pk: True`")
            with pytest.raises(ValidationError):
                parse_from_yaml(StringIO(data))

    def test_bad_mapping_batch_size(self):
        base_path = Path(__file__).parent / "mapping_v2.yml"
        with open(base_path, "r") as f:
            data = f.read().replace("record_type: HH_Account", "batch_size: 50000")
            with pytest.raises(ValidationError):
                parse_from_yaml(StringIO(data))

    def test_ambiguous_mapping_batch_size_default(self, caplog):
        caplog.set_level(logging.WARNING)
        base_path = Path(__file__).parent / "mapping_vanilla_sf.yml"
        with open(base_path, "r") as f:
            data = f.read().replace("table: Opportunity", "batch_size: 150")
            data = data.replace("api: bulk", "")
            parse_from_yaml(StringIO(data))

        assert "If you set a `batch_size` you should also set" in caplog.text

    def test_bad_mapping_batch_size_default(self, caplog):
        caplog.set_level(logging.WARNING)
        base_path = Path(__file__).parent / "mapping_vanilla_sf.yml"
        with open(base_path, "r") as f:
            data = f.read().replace("table: Opportunity", "batch_size: 1000")
            data = f.read().replace("api: bulk", "")
            with pytest.raises(ValidationError):
                parse_from_yaml(StringIO(data))

    def test_default_table_to_sobject_name(self):
        base_path = Path(__file__).parent / "mapping_v3.yml"
        with open(base_path, "r") as f:
            data = f.read()
            ms = parse_from_yaml(StringIO(data))
            assert ms["Insert Accounts"].table == "Account"

    def test_fields_list_to_dict(self):
        base_path = Path(__file__).parent / "mapping_v3.yml"
        with open(base_path, "r") as f:
            data = f.read()
            ms = parse_from_yaml(StringIO(data))
            assert ms["Insert Accounts"].fields == {"Name": "Name"}
            assert ms["Insert Contacts"].fields == {
                "FirstName": "FirstName",
                "LastName": "LastName",
                "Email": "Email",
            }

    def test_fields_default_not_present(self):
        base_path = Path(__file__).parent / "mapping_v3.yml"
        with open(base_path, "r") as f:
            data = f.read()
            ms = parse_from_yaml(StringIO(data))
            assert ms["Insert Junction Objects"].fields == {}

    def test_fields_default_null(self):
        base_path = Path(__file__).parent / "mapping_v3.yml"
        with open(base_path, "r") as f:
            data = f.read()
            ms = parse_from_yaml(StringIO(data))
            assert ms["Insert Other Junction Objects"].fields == {}

    def test_load_from_bytes_stream(self):
        base_path = Path(__file__).parent / "mapping_v2.yml"
        with open(base_path, "rb") as f:
            assert parse_from_yaml(f)

    def test_get_complete_field_map(self):
        m = MappingStep(
            sf_object="Account",
            fields=["Name", "AccountSite"],
            lookups={"ParentId": MappingLookup(table="Account")},
        )

        assert m.get_complete_field_map() == {
            "Name": "Name",
            "AccountSite": "AccountSite",
            "ParentId": "ParentId",
        }
        assert m.get_complete_field_map(include_id=True) == {
            "Id": "sf_id",
            "Name": "Name",
            "AccountSite": "AccountSite",
            "ParentId": "ParentId",
        }

    def test_get_relative_date_context(self):
        mapping = MappingStep(
            sf_object="Account",
            fields=["Some_Date__c", "Some_Datetime__c"],
            anchor_date="2020-07-01",
        )

        salesforce_client = mock.Mock()
        salesforce_client.Account.describe.return_value = {
            "fields": [
                {"name": "Some_Date__c", "type": "date"},
                {"name": "Some_Datetime__c", "type": "datetime"},
                {"name": "Some_Bool__c", "type": "boolean"},
            ]
        }

        assert mapping.get_relative_date_context(
            mapping.get_load_field_list(), salesforce_client
        ) == ([0], [1], date.today())

    def test_get_relative_date_e2e(self):
        base_path = Path(__file__).parent / "mapping_v1.yml"
        mapping = parse_from_yaml(base_path)
        salesforce_client = mock.Mock()
        salesforce_client.Contact.describe.return_value = {
            "fields": [
                {"name": "Some_Date__c", "type": "date"},
                {"name": "Some_Datetime__c", "type": "datetime"},
                {"name": "Some_Bool__c", "type": "boolean"},
            ]
        }
        contacts_mapping = mapping["Insert Contacts"]
        contacts_mapping.fields.update(
            {"Some_Date__c": "Some_Date__c", "Some_Datetime__c": "Some_Datetime__c"}
        )
        assert contacts_mapping.get_relative_date_context(
            contacts_mapping.get_load_field_list(), salesforce_client
        ) == (
            [3],
            [4],
            date.today(),
        )

    # Start of FLS/Namespace Injection Unit Tests

    def test_is_injectable(self):
        assert MappingStep._is_injectable("Test__c")
        assert not MappingStep._is_injectable("npsp__Test__c")
        assert not MappingStep._is_injectable("Account")

    def test_get_permission_types(self):
        ms = MappingStep(
            sf_object="Account", fields=["Name"], action=DataOperationType.INSERT
        )
        assert ms._get_required_permission_types(DataOperationType.INSERT) == (
            "createable",
        )
        assert ms._get_required_permission_types(DataOperationType.QUERY) == (
            "queryable",
        )

        ms = MappingStep(
            sf_object="Account", fields=["Name"], action=DataOperationType.UPDATE
        )
        assert ms._get_required_permission_types(DataOperationType.INSERT) == (
            "updateable",
        )

        ms = MappingStep(
            sf_object="Account",
            fields=["Name", "Extid__c"],
            action=DataOperationType.UPSERT,
            update_key="Extid__c",
        )
        assert ms._get_required_permission_types(DataOperationType.UPSERT) == (
            "updateable",
            "createable",
        )

    def test_check_field_permission(self):
        ms = MappingStep(
            sf_object="Account", fields=["Name"], action=DataOperationType.INSERT
        )

        assert ms._check_field_permission(
            {"Name": {"createable": True}}, "Name", DataOperationType.INSERT
        )

        assert ms._check_field_permission(
            {"Name": {"createable": True}}, "Name", DataOperationType.QUERY
        )

        ms = MappingStep(
            sf_object="Account", fields=["Name"], action=DataOperationType.UPDATE
        )

        assert not ms._check_field_permission(
            {"Name": {"updateable": False}}, "Name", DataOperationType.INSERT
        )

        assert not ms._check_field_permission(
            {"Name": {"updateable": False}}, "Website", DataOperationType.INSERT
        )

    def test_validate_field_dict__fls_checks(self):
        ms = MappingStep(
            sf_object="Account",
            fields=["Id", "Name", "Website"],
            action=DataOperationType.INSERT,
        )

        assert ms._validate_field_dict(
            describe=CaseInsensitiveDict(
                {"Name": {"createable": True}, "Website": {"createable": True}}
            ),
            field_dict=ms.fields_,
            inject=None,
            strip=None,
            drop_missing=False,
            data_operation_type=DataOperationType.INSERT,
        )

        assert not ms._validate_field_dict(
            describe=CaseInsensitiveDict(
                {"Name": {"createable": True}, "Website": {"createable": False}}
            ),
            field_dict=ms.fields_,
            inject=None,
            strip=None,
            drop_missing=False,
            data_operation_type=DataOperationType.INSERT,
        )

    def test_validate_field_dict__injection(self):
        ms = MappingStep(
            sf_object="Account",
            fields=["Id", "Name", "Test__c"],
            action=DataOperationType.INSERT,
        )

        assert ms._validate_field_dict(
            describe=CaseInsensitiveDict(
                {"Name": {"createable": True}, "npsp__Test__c": {"createable": True}}
            ),
            field_dict=ms.fields_,
            inject=lambda field: f"npsp__{field}",
            strip=None,
            drop_missing=False,
            data_operation_type=DataOperationType.INSERT,
        )

        assert ms.fields_ == {"Id": "Id", "Name": "Name", "npsp__Test__c": "Test__c"}

    def test_validate_fields_required(self):
        ms = MappingStep(
            sf_object="Account",
            fields=["Id", "Name", "Test__c"],
            action=DataOperationType.INSERT,
        )
        fields_describe = CaseInsensitiveDict(
            {
                "Name": {
                    "createable": True,
                    "nillable": False,
                    "defaultedOnCreate": False,
                    "defaultValue": None,
                },
                "npsp__Test__c": {
                    "createable": True,
                    "nillable": False,
                    "defaultedOnCreate": False,
                    "defaultValue": None,
                },
            }
        )
        ms._validate_field_dict(
            describe=fields_describe,
            field_dict=ms.fields_,
            inject=lambda field: f"npsp__{field}",
            strip=None,
            drop_missing=False,
            data_operation_type=DataOperationType.INSERT,
        )
        assert ms.fields_ == {"Id": "Id", "Name": "Name", "npsp__Test__c": "Test__c"}
        assert ms.check_required(fields_describe)

        def test_validate_fields_required_missing(self):
            ms = MappingStep(
                sf_object="Account",
                fields=["Test__c"],
                action=DataOperationType.INSERT,
            )
            fields_describe = CaseInsensitiveDict(
                {
                    "Name": {
                        "createable": True,
                        "nillable": False,
                        "defaultedOnCreate": False,
                        "defaultValue": None,
                    },
                    "Test__c": {
                        "createable": True,
                        "nillable": False,
                        "defaultedOnCreate": False,
                        "defaultValue": None,
                    },
                }
            )
            assert ms.fields_ == {"Test__c": "Test__c"}
            assert not ms.check_required(fields_describe)

    def test_validate_field_dict__injection_duplicate_fields(self):
        ms = MappingStep(
            sf_object="Account",
            fields=["Id", "Name", "Test__c"],
            action=DataOperationType.INSERT,
        )

        assert ms._validate_field_dict(
            describe=CaseInsensitiveDict(
                {
                    "Name": {"createable": True},
                    "npsp__Test__c": {"createable": True},
                    "Test__c": {"createable": True},
                }
            ),
            field_dict=ms.fields_,
            inject=lambda field: f"npsp__{field}",
            strip=None,
            drop_missing=False,
            data_operation_type=DataOperationType.INSERT,
        )

        assert ms.fields_ == {"Id": "Id", "Name": "Name", "Test__c": "Test__c"}

    def test_validate_field_dict__drop_missing(self):
        ms = MappingStep(
            sf_object="Account",
            fields=["Id", "Name", "Website"],
            action=DataOperationType.INSERT,
        )

        assert ms._validate_field_dict(
            describe=CaseInsensitiveDict(
                {"Name": {"createable": True}, "Website": {"createable": False}}
            ),
            field_dict=ms.fields_,
            inject=None,
            strip=None,
            drop_missing=True,
            data_operation_type=DataOperationType.INSERT,
        )

        assert ms.fields_ == {"Id": "Id", "Name": "Name"}

    def test_validate_sobject(self):
        ms = MappingStep(
            sf_object="Account", fields=["Name"], action=DataOperationType.INSERT
        )

        assert ms._validate_sobject(
            CaseInsensitiveDict({"Account": {"createable": True}}),
            None,
            None,
            DataOperationType.INSERT,
        )

        assert ms._validate_sobject(
            CaseInsensitiveDict({"Account": {"queryable": True}}),
            None,
            None,
            DataOperationType.QUERY,
        )

        ms = MappingStep(
            sf_object="Account", fields=["Name"], action=DataOperationType.UPDATE
        )

        assert not ms._validate_sobject(
            CaseInsensitiveDict({"Account": {"updateable": False}}),
            None,
            None,
            DataOperationType.INSERT,
        )

    def test_validate_sobject__injection(self):
        ms = MappingStep(
            sf_object="Test__c", fields=["Name"], action=DataOperationType.INSERT
        )

        assert ms._validate_sobject(
            CaseInsensitiveDict({"npsp__Test__c": {"createable": True}}),
            inject=lambda obj: f"npsp__{obj}",
            strip=None,
            data_operation_type=DataOperationType.INSERT,
        )
        assert ms.sf_object == "npsp__Test__c"

    def test_validate_sobject__stripping(self):
        ms = MappingStep(
            sf_object="foo__Test__c", fields=["Name"], action=DataOperationType.INSERT
        )

        assert ms._validate_sobject(
            CaseInsensitiveDict({"Test__c": {"createable": True}}),
            inject=None,
            strip=lambda obj: obj[len("foo__") :],
            data_operation_type=DataOperationType.INSERT,
        )
        assert ms.sf_object == "Test__c"

    def test_validate_sobject__injection_duplicate(self):
        ms = MappingStep(
            sf_object="Test__c", fields=["Name"], action=DataOperationType.INSERT
        )

        assert ms._validate_sobject(
            CaseInsensitiveDict(
                {"npsp__Test__c": {"createable": True}, "Test__c": {"createable": True}}
            ),
            lambda obj: f"npsp__{obj}",
            None,
            DataOperationType.INSERT,
        )
        assert ms.sf_object == "Test__c"

    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_sobject",
        return_value=True,
    )
    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_field_dict",
        return_value=True,
    )
    def test_validate_and_inject_namespace__injection_fields(
        self, mock_field, mock_sobject
    ):
        ms = parse_from_yaml(
            StringIO(
                """Insert Accounts:
                  sf_object: Account
                  table: Account
                  fields:
                    - Test__c"""
            )
        )["Insert Accounts"]

        salesforce_client = mock.Mock()
        salesforce_client.describe.return_value = {
            "sobjects": [{"name": "Account", "createable": True}]
        }
        salesforce_client.Account.describe.return_value = {
            "fields": [{"name": "ns__Test__c", "createable": True}]
        }

        assert ms.validate_and_inject_namespace(
            salesforce_client, "ns", DataOperationType.INSERT, inject_namespaces=True
        )

        ms._validate_sobject.assert_called_once_with(
            CaseInsensitiveDict({"Account": {"name": "Account", "createable": True}}),
            mock.ANY,  # This is a function def
            mock.ANY,
            DataOperationType.INSERT,
        )

        ms._validate_field_dict.assert_has_calls(
            [
                mock.call(
                    CaseInsensitiveDict(
                        {"ns__Test__c": {"name": "ns__Test__c", "createable": True}}
                    ),
                    ms.fields,
                    mock.ANY,  # local function def
                    mock.ANY,  # local function def
                    False,
                    DataOperationType.INSERT,
                ),
                mock.call(
                    {"ns__Test__c": {"name": "ns__Test__c", "createable": True}},
                    ms.lookups,
                    mock.ANY,  # local function def
                    mock.ANY,  # local function def
                    False,
                    DataOperationType.INSERT,
                ),
            ]
        )

    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_sobject",
        return_value=True,
    )
    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_field_dict",
        return_value=True,
    )
    def test_validate_and_inject_namespace__injection_lookups(
        self, mock_field, mock_sobject
    ):
        ms = parse_from_yaml(
            StringIO(
                """Insert Accounts:
                  sf_object: Account
                  table: Account
                  fields:
                    - Name
                  lookups:
                    Lookup__c:
                        table: Stuff"""
            )
        )["Insert Accounts"]

        salesforce_client = mock.Mock()
        salesforce_client.describe.return_value = {
            "sobjects": [{"name": "Account", "createable": True}]
        }
        salesforce_client.Account.describe.return_value = {
            "fields": [
                {"name": "Name", "createable": True},
                {"name": "ns__Lookup__c", "updateable": False, "createable": True},
            ]
        }

        assert ms.validate_and_inject_namespace(
            salesforce_client, "ns", DataOperationType.INSERT, inject_namespaces=True
        )

        ms._validate_sobject.assert_called_once_with(
            CaseInsensitiveDict({"Account": {"name": "Account", "createable": True}}),
            mock.ANY,  # local function def
            mock.ANY,
            DataOperationType.INSERT,
        )

        ms._validate_field_dict.assert_has_calls(
            [
                mock.call(
                    {
                        "Name": {"name": "Name", "createable": True},
                        "ns__Lookup__c": {
                            "name": "ns__Lookup__c",
                            "updateable": False,
                            "createable": True,
                        },
                    },
                    ms.fields,
                    mock.ANY,  # local function def.
                    mock.ANY,  # local function def.
                    False,
                    DataOperationType.INSERT,
                ),
                mock.call(
                    {
                        "Name": {"name": "Name", "createable": True},
                        "ns__Lookup__c": {
                            "name": "ns__Lookup__c",
                            "updateable": False,
                            "createable": True,
                        },
                    },
                    ms.lookups,
                    mock.ANY,  # local function def.
                    mock.ANY,  # local function def.
                    False,
                    DataOperationType.INSERT,
                ),
            ]
        )

    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_sobject",
        return_value=True,
    )
    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_field_dict",
        return_value=True,
    )
    def test_validate_and_inject_namespace__fls(self, mock_field, mock_sobject):
        ms = MappingStep(
            sf_object="Test__c", fields=["Field__c"], action=DataOperationType.INSERT
        )

        salesforce_client = mock.Mock()
        salesforce_client.describe.return_value = {
            "sobjects": [{"name": "Test__c", "createable": True}]
        }
        salesforce_client.Test__c.describe.return_value = {
            "fields": [{"name": "Field__c", "createable": True}]
        }
        assert ms.validate_and_inject_namespace(
            salesforce_client, "ns", DataOperationType.INSERT
        )

        ms._validate_sobject.assert_called_once_with(
            CaseInsensitiveDict({"Test__c": {"name": "Test__c", "createable": True}}),
            None,
            None,
            DataOperationType.INSERT,
        )

        ms._validate_field_dict.assert_has_calls(
            [
                mock.call(
                    {"Field__c": {"name": "Field__c", "createable": True}},
                    {"Field__c": "Field__c"},
                    None,
                    None,
                    False,
                    DataOperationType.INSERT,
                ),
                mock.call(
                    {"Field__c": {"name": "Field__c", "createable": True}},
                    {},
                    None,
                    None,
                    False,
                    DataOperationType.INSERT,
                ),
            ]
        )

    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_sobject",
        return_value=False,
    )
    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_field_dict",
        return_value=True,
    )
    def test_validate_and_inject_namespace__fls_sobject_failure(
        self, mock_field, mock_sobject
    ):
        ms = MappingStep(
            sf_object="Test__c", fields=["Name"], action=DataOperationType.INSERT
        )

        salesforce_client = mock.Mock()
        salesforce_client.describe.return_value = {
            "sobjects": [{"name": "Test__c", "createable": False}]
        }
        salesforce_client.Test__c.describe.return_value = {
            "fields": [{"name": "Name", "createable": True}]
        }
        assert not ms.validate_and_inject_namespace(
            salesforce_client, "ns", DataOperationType.INSERT
        )

        ms._validate_sobject.assert_called_once_with(
            {"Test__c": {"name": "Test__c", "createable": False}},
            None,
            None,
            DataOperationType.INSERT,
        )

        ms._validate_field_dict.assert_not_called()

    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_sobject",
        return_value=True,
    )
    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_field_dict",
        return_value=False,
    )
    def test_validate_and_inject_namespace__fls_fields_failure(
        self, mock_field, mock_sobject
    ):
        ms = MappingStep(
            sf_object="Test__c", fields=["Name"], action=DataOperationType.INSERT
        )

        salesforce_client = mock.Mock()
        salesforce_client.describe.return_value = {
            "sobjects": [{"name": "Test__c", "createable": True}]
        }
        salesforce_client.Test__c.describe.return_value = {
            "fields": [{"name": "Name", "createable": False}]
        }
        assert not ms.validate_and_inject_namespace(
            salesforce_client, "ns", DataOperationType.INSERT
        )

        ms._validate_sobject.assert_called_once_with(
            {"Test__c": {"name": "Test__c", "createable": True}},
            None,
            None,
            DataOperationType.INSERT,
        )

        ms._validate_field_dict.assert_has_calls(
            [
                mock.call(
                    {"Name": {"name": "Name", "createable": False}},
                    {"Name": "Name"},
                    None,
                    None,
                    False,
                    DataOperationType.INSERT,
                )
            ]
        )

    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_sobject",
        return_value=True,
    )
    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_field_dict",
        side_effect=[True, False],
    )
    def test_validate_and_inject_namespace__fls_lookups_failure(
        self, mock_field, mock_sobject
    ):
        ms = parse_from_yaml(
            StringIO(
                """Insert Accounts:
                  sf_object: Account
                  table: Account
                  fields:
                    - Name
                  lookups:
                    Lookup__c:
                        table: Stuff"""
            )
        )["Insert Accounts"]

        salesforce_client = mock.Mock()
        salesforce_client.describe.return_value = {
            "sobjects": [{"name": "Account", "createable": True}]
        }
        salesforce_client.Account.describe.return_value = {
            "fields": [
                {"name": "Name", "createable": True},
                {"name": "Lookup__c", "updateable": True, "createable": False},
            ]
        }
        assert not ms.validate_and_inject_namespace(
            salesforce_client, "ns", DataOperationType.INSERT
        )

        ms._validate_sobject.assert_called_once_with(
            {"Account": {"name": "Account", "createable": True}},
            None,
            None,
            DataOperationType.INSERT,
        )

        ms._validate_field_dict.assert_has_calls(
            [
                mock.call(
                    {
                        "Name": {"name": "Name", "createable": True},
                        "Lookup__c": {
                            "name": "Lookup__c",
                            "updateable": True,
                            "createable": False,
                        },
                    },
                    {"Name": "Name"},
                    None,
                    None,
                    False,
                    DataOperationType.INSERT,
                ),
                mock.call(
                    {
                        "Name": {"name": "Name", "createable": True},
                        "Lookup__c": {
                            "name": "Lookup__c",
                            "updateable": True,
                            "createable": False,
                        },
                    },
                    ms.lookups,
                    None,
                    None,
                    False,
                    DataOperationType.INSERT,
                ),
            ]
        )

    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_sobject",
        return_value=True,
    )
    @mock.patch(
        "cumulusci.tasks.bulkdata.mapping_parser.MappingStep._validate_field_dict",
        side_effect=[True, False],
    )
    def test_validate_and_inject_namespace__fls_lookups_update_failure(
        self, mock_field, mock_sobject
    ):
        ms = parse_from_yaml(
            StringIO(
                """Insert Accounts:
                  sf_object: Account
                  table: Account
                  fields:
                    - Name
                  lookups:
                    Lookup__c:
                        table: Stuff
                        after: Insert Stuff"""
            )
        )["Insert Accounts"]

        salesforce_client = mock.Mock()
        salesforce_client.describe.return_value = {
            "sobjects": [{"name": "Account", "createable": True}]
        }
        salesforce_client.Account.describe.return_value = {
            "fields": [
                {"name": "Name", "createable": True},
                {"name": "Lookup__c", "updateable": False, "createable": True},
            ]
        }
        assert not ms.validate_and_inject_namespace(
            salesforce_client, "ns", DataOperationType.INSERT
        )

        ms._validate_sobject.assert_called_once_with(
            {"Account": {"name": "Account", "createable": True}},
            None,
            None,
            DataOperationType.INSERT,
        )

        ms._validate_field_dict.assert_has_calls(
            [
                mock.call(
                    {
                        "Name": {"name": "Name", "createable": True},
                        "Lookup__c": {
                            "name": "Lookup__c",
                            "updateable": False,
                            "createable": True,
                        },
                    },
                    {"Name": "Name"},
                    None,
                    None,
                    False,
                    DataOperationType.INSERT,
                ),
                mock.call(
                    {
                        "Name": {"name": "Name", "createable": True},
                        "Lookup__c": {
                            "name": "Lookup__c",
                            "updateable": False,
                            "createable": True,
                        },
                    },
                    ms.lookups,
                    None,
                    None,
                    False,
                    DataOperationType.INSERT,
                ),
            ]
        )

    # Start of FLS/Namespace Injection Integration Tests

    @responses.activate
    def test_validate_and_inject_mapping_enforces_fls(self):
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                "Insert Accounts:\n  sf_object: Account\n  table: Account\n  fields:\n    - Nonsense__c"
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        with pytest.raises(BulkDataException):
            validate_and_inject_mapping(
                mapping=mapping,
                sf=org_config.salesforce_client,
                namespace=None,
                data_operation=DataOperationType.INSERT,
                inject_namespaces=False,
                drop_missing=False,
            )

    @responses.activate
    def test_validate_and_inject_mapping_removes_steps_with_drop_missing(self):
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                "Insert Accounts:\n  sf_object: NotAccount\n  table: Account\n  fields:\n    - Nonsense__c"
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        validate_and_inject_mapping(
            mapping=mapping,
            sf=org_config.salesforce_client,
            namespace=None,
            data_operation=DataOperationType.INSERT,
            inject_namespaces=False,
            drop_missing=True,
        )

        assert "Insert Accounts" not in mapping

    @responses.activate
    def test_validate_and_inject_mapping_removes_lookups_with_drop_missing(self):
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Accounts:\n  sf_object: NotAccount\n  table: Account\n  fields:\n    - Nonsense__c\n"
                    "Insert Contacts:\n  sf_object: Contact\n  table: Contact\n  fields:\n    - LastName\n  lookups:\n    AccountId:\n      table: Account"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        validate_and_inject_mapping(
            mapping=mapping,
            sf=org_config.salesforce_client,
            namespace=None,
            data_operation=DataOperationType.INSERT,
            inject_namespaces=False,
            drop_missing=True,
        )

        assert "Insert Accounts" not in mapping
        assert "Insert Contacts" in mapping
        assert "AccountId" not in mapping["Insert Contacts"].lookups

    @responses.activate
    def test_validate_and_inject_mapping_removes_lookups_with_drop_missing__polymorphic_partial_present(
        self,
    ):
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Accounts:\n  sf_object: NotAccount\n  table: Account\n  fields:\n    - Nonsense__c\n"
                    "Insert Contacts:\n  sf_object: Contact\n  table: Contact\n  lookups:\n    AccountId:\n      table: Account\n"
                    "Insert Events:\n  sf_object: Event\n  table: Event\n  lookups:\n    WhoId:\n      table:\n        - Contact\n        - Lead"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        validate_and_inject_mapping(
            mapping=mapping,
            sf=org_config.salesforce_client,
            namespace=None,
            data_operation=DataOperationType.QUERY,
            inject_namespaces=False,
            drop_missing=True,
        )

        assert "Insert Accounts" not in mapping
        assert "Insert Contacts" in mapping
        assert "Insert Events" in mapping
        assert "AccountId" not in mapping["Insert Contacts"].lookups
        assert "WhoId" in mapping["Insert Events"].lookups

    @responses.activate
    def test_validate_and_inject_mapping_removes_lookups_with_drop_missing__polymorphic_none_present(
        self,
    ):
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Contacts:\n  sf_object: NotContact\n  table: NotContact\n  fields:\n    - LastName\n"
                    "Insert Leads:\n  sf_object: NotLead\n  table: NotLead\n  fields:\n    - LastName\n    - Company\n"
                    "Insert Events:\n  sf_object: Event\n  table: Event\n  lookups:\n    WhoId:\n      table:\n        - Contact\n        - Lead"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        validate_and_inject_mapping(
            mapping=mapping,
            sf=org_config.salesforce_client,
            namespace=None,
            data_operation=DataOperationType.QUERY,
            inject_namespaces=False,
            drop_missing=True,
        )

        assert "Insert Contacts" not in mapping
        assert "Insert Leads" not in mapping
        assert "Insert Events" in mapping
        assert "WhoId" not in mapping["Insert Events"].lookups

    @responses.activate
    def test_validate_and_inject_mapping_throws_exception_required_lookup_dropped(self):
        mock_describe_calls()

        # This test uses a bit of gimmickry to validate exception behavior on dropping a required lookup.
        # Since mapping_parser identifies target objects via the `table` clause rather than the actual schema,
        # which makes us resilient to polymorphic lookups, we'll pretend the non-nillable `Id` field is a lookup.
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Accounts:\n  sf_object: NotAccount\n  table: Account\n  fields:\n    - Nonsense__c\n"
                    "Insert Contacts:\n  sf_object: Contact\n  table: Contact\n  fields:\n    - LastName\n  lookups:\n    Id:\n      table: Account"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        with pytest.raises(BulkDataException):
            validate_and_inject_mapping(
                mapping=mapping,
                sf=org_config.salesforce_client,
                namespace=None,
                data_operation=DataOperationType.INSERT,
                inject_namespaces=False,
                drop_missing=True,
            )

    @responses.activate
    def test_validate_and_inject_mapping_throws_exception_required_fields_missing(
        self, caplog
    ):
        caplog.set_level(logging.ERROR)
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Accounts:\n  sf_object: Account\n  table: Account\n  fields:\n    - ns__Description__c\n"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        validate_and_inject_mapping(
            mapping=mapping,
            sf=org_config.salesforce_client,
            namespace="",
            data_operation=DataOperationType.INSERT,
            inject_namespaces=False,
            drop_missing=False,
        )

        expected_error_message = (
            "One or more required fields are missing for loading on Account :{'Name'}"
        )
        error_logs = [
            record.message for record in caplog.records if record.levelname == "ERROR"
        ]
        assert any(expected_error_message in error_log for error_log in error_logs)

    @responses.activate
    def test_validate_and_inject_mapping_injects_namespaces(self):
        mock_describe_calls()
        # Note: ns__Description__c is a mock field added to our stored, mock describes (in JSON)
        ms = parse_from_yaml(
            StringIO(
                """Insert Accounts:
                  sf_object: Account
                  table: Account
                  fields:
                    - Description__c"""
            )
        )["Insert Accounts"]
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        assert ms.validate_and_inject_namespace(
            org_config.salesforce_client,
            "ns",
            DataOperationType.INSERT,
            inject_namespaces=True,
        )

        assert list(ms.fields.keys()) == ["ns__Description__c"]

    @responses.activate
    def test_validate_and_inject_mapping_injects_namespaces__validates_lookup(self):
        """Test to verify that with namespace inject, we validate lookups correctly"""
        mock_describe_calls()
        # Note: ns__Description__c is a mock field added to our stored, mock describes (in JSON)
        mapping = parse_from_yaml(
            StringIO(
                """Insert Accounts:
                  sf_object: Account
                  table: Account
                  fields:
                    - Description__c
                  lookups:
                    LinkedAccount__c:
                      table: Account"""
            )
        )
        ms = mapping["Insert Accounts"]
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        assert ms.validate_and_inject_namespace(
            org_config.salesforce_client,
            "ns",
            DataOperationType.INSERT,
            inject_namespaces=True,
        )
        # Here we verify that the field ns__LinkedAccount__c does lookup
        # to sobject Account inside of describe
        _infer_and_validate_lookups(mapping, org_config.salesforce_client)
        assert list(ms.lookups.keys()) == ["ns__LinkedAccount__c"]

    @responses.activate
    def test_validate_and_inject_mapping_removes_namespaces(self):
        mock_describe_calls()
        # Note: History__c is a mock field added to our stored, mock describes (in JSON)
        ms = parse_from_yaml(
            StringIO(
                """Insert Accounts:
                  sf_object: Account
                  table: Account
                  fields:
                    - ns__History__c"""
            )
        )["Insert Accounts"]
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        assert ms.validate_and_inject_namespace(
            org_config.salesforce_client,
            "ns",
            DataOperationType.INSERT,
            inject_namespaces=True,
        )

        assert list(ms.fields.keys()) == ["History__c"]

    @responses.activate
    def test_validate_and_inject_mapping_queries_is_person_account_field(self):
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Accounts:\n  sf_object: Account\n  table: Account\n  fields:\n    - Description__c\n"
                    "Insert Contacts:\n  sf_object: Contact\n  table: Contact\n  lookups:\n    AccountId:\n      table: Account"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        validate_and_inject_mapping(
            mapping=mapping,
            sf=org_config.salesforce_client,
            namespace=None,
            data_operation=DataOperationType.QUERY,
            inject_namespaces=False,
            drop_missing=True,
            org_has_person_accounts_enabled=True,
        )

        assert "Insert Accounts" in mapping
        assert "Insert Contacts" in mapping
        assert "IsPersonAccount" in mapping["Insert Accounts"]["fields"]
        assert "IsPersonAccount" in mapping["Insert Contacts"]["fields"]


class TestMappingLookup:
    def test_get_lookup_key_field__no_model(self):
        lookup = MappingLookup(table="contact", name="AccountId")
        assert lookup.get_lookup_key_field() == "AccountId"

    def test_get_lookup_key_field__snake_case_model(self):
        class FakeModel:
            account_id = mock.MagicMock()

        lookup = MappingLookup(table="contact", name="AccountId")
        assert lookup.get_lookup_key_field(FakeModel()) == "account_id"

    def test_get_lookup_key_field__by_key_field(self):
        class FakeModel:
            foo = mock.MagicMock()

        lookup = MappingLookup(table="contact", key_field="foo", name="AccountId")
        assert lookup.get_lookup_key_field(FakeModel()) == "foo"

    def test_get_lookup_key_field__by_key_field_wrong_case(self):
        class FakeModel:
            account_id = mock.MagicMock()

        # we can correct mismatched mapping files if the mistake is just
        # old-fashioned SQL with new Mapping File
        lookup = MappingLookup(table="contact", key_field="AccountId", name="AccountId")
        assert lookup.get_lookup_key_field(FakeModel()) == "account_id"

    def test_get_lookup_key_field__mismatched_name(self):
        class FakeModel:
            account_id = mock.MagicMock()

        # some mistakes can't be fixed.
        lookup = MappingLookup(table="contact", key_field="Foo", name="Foo")

        with pytest.raises(KeyError):
            lookup.get_lookup_key_field(FakeModel())

    @responses.activate
    def test_validate_and_inject_mapping_works_case_insensitively(self):
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Accounts:\n  sf_object: account\n  table: account\n  fields:\n    - name\n"
                    "Insert Contacts:\n  sf_object: contact\n  table: contact\n  fields:\n    - LaSTnAME\n  lookups:\n    accountid:\n      table: account"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        assert mapping["Insert Accounts"].sf_object != "Account"
        assert mapping["Insert Accounts"].sf_object == "account"
        assert "name" in mapping["Insert Accounts"].fields
        assert "Name" not in mapping["Insert Accounts"].fields

        validate_and_inject_mapping(
            mapping=mapping,
            sf=org_config.salesforce_client,
            namespace=None,
            data_operation=DataOperationType.INSERT,
            inject_namespaces=False,
            drop_missing=False,
        )
        assert mapping["Insert Accounts"].sf_object == "Account"
        assert mapping["Insert Accounts"].sf_object != "account"
        assert "Name" in mapping["Insert Accounts"].fields
        assert "name" not in mapping["Insert Accounts"].fields

    @responses.activate
    def test_bulk_attributes(self):
        mapping = parse_from_yaml(
            StringIO(
                (
                    """Insert Accounts:
                        sf_object: account
                        table: account
                        api: rest
                        bulk_mode: Serial
                        batch_size: 50
                        fields:
                            - name"""
                )
            )
        )
        assert mapping["Insert Accounts"].api == DataApi.REST
        assert mapping["Insert Accounts"].bulk_mode == "Serial"
        assert mapping["Insert Accounts"].batch_size == 50

    def test_case_conversions(self):
        mapping = parse_from_yaml(
            StringIO(
                (
                    """Insert Accounts:
                        sf_object: account
                        table: account
                        api: ReST
                        bulk_mode: serial
                        action: INSerT
                        batch_size: 50
                        fields:
                            - name"""
                )
            )
        )
        assert mapping["Insert Accounts"].api == DataApi.REST
        assert mapping["Insert Accounts"].bulk_mode == "Serial"
        assert mapping["Insert Accounts"].action.value == "insert"
        assert mapping["Insert Accounts"].batch_size == 50

    def test_oid_as_pk__raises(self):
        with pytest.raises(ValueError):
            parse_from_yaml(
                StringIO(
                    (
                        """Insert Accounts:
                            sf_object: account
                            oid_as_pk: True
                            fields:
                                - name"""
                    )
                )
            )

    def test_oid_as_pk__false(self):
        result = parse_from_yaml(
            StringIO(
                (
                    """Insert Accounts:
                            sf_object: account
                            oid_as_pk: False
                            fields:
                                - name"""
                )
            )
        )
        assert result["Insert Accounts"].oid_as_pk is False


class TestUpsertKeyValidations:
    def test_upsert_key_wrong_type(self):
        with pytest.raises(ValidationError) as e:
            parse_from_yaml(
                StringIO(
                    (
                        """Insert Accounts:
                        sf_object: account
                        table: account
                        action: upsert
                        update_key: 11
                        fields:
                            - name"""
                    )
                )
            )
        assert "update_key" in str(e.value)

    def test_upsert_key_wrong_type__list_item(self):
        with pytest.raises(ValidationError) as e:
            parse_from_yaml(
                StringIO(
                    (
                        """Insert Accounts:
                        sf_object: account
                        table: account
                        action: upsert
                        update_key:
                            - 11
                        fields:
                            - name"""
                    )
                )
            )
        assert "update_key" in str(e.value)

    def test_upsert_key_list(self):
        mapping = parse_from_yaml(
            StringIO(
                (
                    """Insert Accounts:
                        sf_object: account
                        table: account
                        action: etl_upsert
                        update_key:
                            - FirstName
                            - LastName
                        fields:
                            - FirstName
                            - LastName """
                )
            )
        )
        assert mapping["Insert Accounts"]["update_key"] == (
            "FirstName",
            "LastName",
        ), mapping["Insert Accounts"]["update_key"]

    def test_get_extract_field_list(self):
        """Test to ensure Id comes first, lookups come
        last and order of mappings do not change"""
        m = MappingStep(
            sf_object="Account",
            fields=["Name", "RecordTypeId", "AccountSite"],
            lookups={"ParentId": MappingLookup(table="Account")},
        )

        assert m.get_extract_field_list() == [
            "Id",
            "Name",
            "RecordTypeId",
            "AccountSite",
            "ParentId",
        ]

    @responses.activate
    def test_infer_and_validate_lookups__table_doesnt_exist(self, caplog):
        caplog.set_level(logging.ERROR)
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Contacts:\n  sf_object: Contact\n  table: Contact\n  lookups:\n    AccountId:\n      table: Account"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        expected_error_message = "The table Account does not exist in the mapping file"

        with pytest.raises(BulkDataException) as e:
            _infer_and_validate_lookups(
                mapping=mapping, sf=org_config.salesforce_client
            )
            assert "One or more relationship errors blocked the operation" in str(
                e.value
            )
        error_logs = [
            record.message for record in caplog.records if record.levelname == "ERROR"
        ]
        assert any(expected_error_message in error_log for error_log in error_logs)

    @responses.activate
    def test_infer_and_validate_lookups__incorrect_order(self, caplog):
        caplog.set_level(logging.ERROR)
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Account:\n"
                    "  sf_object: Account\n"
                    "  table: Account\n"
                    "  fields:\n"
                    "    - Description__c\n"
                    "Insert Events:\n"
                    "  sf_object: Event\n"
                    "  table: Event\n"
                    "  lookups:\n"
                    "    WhatId:\n"
                    "      table:\n"
                    "        - Opportunity\n"
                    "        - Account\n"
                    "Insert Opportunity:\n"
                    "  sf_object: Opportunity\n"
                    "  table: Opportunity\n"
                    "  fields:\n"
                    "    - Description__c\n"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        expected_error_message = "All included target objects (Opportunity,Account) for the field Event.WhatId must precede Event in the mapping."

        with pytest.raises(BulkDataException) as e:
            _infer_and_validate_lookups(
                mapping=mapping, sf=org_config.salesforce_client
            )
            assert "One or more relationship errors blocked the operation" in str(
                e.value
            )
        error_logs = [
            record.message for record in caplog.records if record.levelname == "ERROR"
        ]
        assert any(expected_error_message in error_log for error_log in error_logs)

    @responses.activate
    def test_infer_and_validate_lookups__after(self):
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Accounts:\n  sf_object: Account\n  table: Account\n  lookups:\n    ParentId:\n      table: Account\n"
                    "Insert Contacts:\n  sf_object: Contact\n  table: Contact\n  lookups:\n    AccountId:\n      table: Account"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )
        _infer_and_validate_lookups(mapping=mapping, sf=org_config.salesforce_client)
        assert mapping["Insert Accounts"].lookups["ParentId"].after == "Insert Accounts"
        assert mapping["Insert Contacts"].lookups["AccountId"].after is None

    @responses.activate
    def test_infer_and_validate_lookups__invalid_reference(self, caplog):
        caplog.set_level(logging.ERROR)
        mock_describe_calls()
        mapping = parse_from_yaml(
            StringIO(
                (
                    "Insert Events:\n  sf_object: Event\n  table: Event\n  fields:\n    - Description__c\n"
                    "Insert Contacts:\n  sf_object: Contact\n  table: Contact\n  lookups:\n    AccountId:\n      table: Event"
                )
            )
        )
        org_config = DummyOrgConfig(
            {"instance_url": "https://example.com", "access_token": "abc123"}, "test"
        )

        expected_error_message = "The lookup Event is not a valid lookup"

        with pytest.raises(BulkDataException) as e:
            _infer_and_validate_lookups(
                mapping=mapping, sf=org_config.salesforce_client
            )
            assert "One or more relationship errors blocked the operation" in str(
                e.value
            )
        error_logs = [
            record.message for record in caplog.records if record.levelname == "ERROR"
        ]
        assert any(expected_error_message in error_log for error_log in error_logs)



import contextlib
import json
import re
from pathlib import Path
from typing import List, Optional, Union

import requests

from cumulusci.core.config import BaseProjectConfig, FlowConfig, TaskConfig
from cumulusci.core.exceptions import CumulusCIException, TaskOptionsError
from cumulusci.core.flowrunner import FlowCoordinator
from cumulusci.core.github import get_tag_by_name
from cumulusci.core.metadeploy.api import MetaDeployAPI
from cumulusci.core.tasks import BaseTask
from cumulusci.core.utils import process_bool_arg
from cumulusci.utils import cd, download_extract_github, temporary_dir
from cumulusci.utils.http.requests_utils import safe_json_from_response
from cumulusci.utils.yaml.cumulusci_yml import Plan

INSTALL_VERSION_RE = re.compile(r"^Install .*\d$")


class BaseMetaDeployTask(BaseTask):
    """Base class for tasks that talk to MetaDeploy's API."""

    def _init_task(self):
        metadeploy_service = self.project_config.keychain.get_service("metadeploy")
        self.api: MetaDeployAPI = MetaDeployAPI(metadeploy_service)

    def _call_api(self, method, path, collect_pages=False, **kwargs):
        next_url = path
        results = []
        while next_url is not None:
            response = self.api.session.request(method, next_url, **kwargs)
            if response.status_code == 400:
                raise requests.exceptions.HTTPError(response.content)
            response = safe_json_from_response(response)
            if "links" in response and collect_pages:
                results.extend(response["data"])
                next_url = response["links"]["next"]
            else:
                return response
        return results


class Publish(BaseMetaDeployTask):
    """Publishes installation plans to MetaDeploy."""

    task_options = {
        "tag": {"description": "Name of the git tag to publish"},
        "commit": {"description": "Commit hash to publish"},
        "plan": {
            "description": "Name of the plan(s) to publish. "
            "This refers to the `plans` section of cumulusci.yml. "
            "By default, all plans will be published.",
            "required": False,
        },
        "dry_run": {
            "description": "If True, print steps without publishing.",
            "required": False,
        },
        "publish": {
            "description": "If True, set is_listed to True on the version. Default: False",
            "required": False,
        },
        "labels_path": {
            "description": "Path to a folder containing translations.",
            "required": False,
        },
    }

    def _init_task(self):
        super(Publish, self)._init_task()
        self.dry_run = process_bool_arg(self.options.get("dry_run") or False)
        self.publish = not self.dry_run and process_bool_arg(
            self.options.get("publish") or False
        )
        self.tag = self.options.get("tag")
        self.commit = self.options.get("commit")
        if not self.tag and not self.commit:
            raise TaskOptionsError("You must specify either the tag or commit option.")
        self.labels_path = self.options.get("labels_path", "metadeploy")
        Path(self.labels_path).mkdir(parents=True, exist_ok=True)

        if plan_name := self.options.get("plan"):
            if self.project_config.lookup(f"plans__{plan_name}") is None:
                raise TaskOptionsError(
                    f"Plan {plan_name} not found in project configuration"
                )
            plan_configs = {
                plan_name: self.project_config.lookup(f"plans__{plan_name}")
            }
            self.plan_configs = plan_configs
        else:
            self.plan_configs = self.project_config.plans
        # Handled exception for no plan
        if self.plan_configs is None or len(self.plan_configs) == 0:
            raise CumulusCIException(
                "No plan found to publish in project configuration"
            )

        self._load_labels()

    def _run_task(self):
        repo_owner = self.project_config.repo_owner
        repo_name = self.project_config.repo_name
        repo_url = f"https://github.com/{repo_owner}/{repo_name}"

        # Find or create Version
        product = self._find_product(repo_url)
        if not product:
            raise CumulusCIException(
                f"No product found in MetaDeploy with repo URL {repo_url}"
            )
        elif not product.get("slug"):  # pragma: no cover
            raise CumulusCIException(
                f"No slug found in MetaDeploy for product {product} from {repo_url}"
            )
        if not self.dry_run:
            version = self._find_or_create_version(product)
            if self.labels_path and "slug" in product:
                self._publish_labels(product["slug"])

        # Check out the specified tag
        gh = self.project_config.get_github_api()
        repo = gh.repository(repo_owner, repo_name)
        if self.tag:
            tag = get_tag_by_name(repo, self.tag)
            self.commit = tag.object.sha
        self.logger.info(
            f"Downloading commit {self.commit} of {repo.full_name} from GitHub"
        )
        zf = download_extract_github(gh, repo_owner, repo_name, ref=self.commit)
        with temporary_dir() as project_dir:
            zf.extractall(project_dir)
            project_config = BaseProjectConfig(
                self.project_config.universal_config_obj,
                repo_info={
                    "root": project_dir,
                    "owner": repo_owner,
                    "name": repo_name,
                    "url": repo_url,
                    "branch": self.tag or self.commit,
                    "commit": self.commit,
                },
            )
            project_config.set_keychain(self.project_config.keychain)

            for plan_name, plan_config in self.plan_configs.items():
                self._add_plan_labels(
                    plan_name=plan_name,
                    plan_config=plan_config,
                )

                steps = self._freeze_steps(project_config, plan_config)

                if not self.dry_run:
                    self._publish_plan(product, version, plan_name, plan_config, steps)

            # Update version to set is_listed=True
            if self.publish:
                self.api.update_version(version["id"])
                self.logger.info(f"Published Version {version['url']}")

        # Save labels
        self._save_labels()

    def _freeze_steps(self, project_config, plan_config) -> list:
        steps = plan_config["steps"]
        flow_config = FlowConfig(plan_config)
        flow_config.project_config = project_config
        flow = FlowCoordinator(project_config, flow_config)
        steps = []
        for step in flow.steps:
            if step.skip:
                continue
            with cd(step.project_config.repo_root):
                task = step.task_class(
                    step.project_config,
                    TaskConfig(step.task_config),
                    name=step.task_name,
                )
                steps.extend(task.freeze(step))
        self.logger.debug("Prepared steps:\n" + json.dumps(steps, indent=4))

        self._add_step_labels(steps)
        return steps

    def _add_plan_labels(self, plan_name, plan_config) -> None:
        """Add labels for plans, steps, and checks."""
        self._add_labels(
            plan_config,
            f"plan:{plan_name}",
            {
                "title": "title of installation plan",
                "preflight_message": "shown before user starts installation (markdown)",
                "preflight_message_additional": "shown before user starts installation (markdown)",
                "post_install_message": "shown after successful installation (markdown)",
                "post_install_message_additional": "shown after successful installation (markdown)",
                "error_message": "shown after failed installation (markdown)",
            },
        )
        self._add_check_labels(plan_config)

    def _add_step_labels(self, steps) -> None:
        for step in steps:
            # avoid separate labels for installing each package
            if INSTALL_VERSION_RE.match(step["name"]):
                self._add_label(
                    "steps",
                    "Install {product} {version}",
                    "title of installation step",
                )
            else:
                self._add_label("steps", step["name"], "title of installation step")
            self._add_label(
                "steps",
                step.get("description"),
                "description of installation step",
            )
            self._add_check_labels(step["task_config"])

    def _add_check_labels(self, config) -> None:
        for check in config.get("checks", []):
            self._add_label("checks", check.get("message"), "shown if validation fails")

    def _publish_plan(self, product, version, plan_name, plan_config, steps):
        plan_template = self._find_or_create_plan_template(
            product, plan_name, plan_config
        )

        allowed_org_providers = self._get_allowed_org_providers(plan_name)
        supported_orgs = self._convert_org_providers_to_plan_equivalent(
            allowed_org_providers
        )

        plan_json = {
            "is_listed": plan_config.get("is_listed", True),
            "plan_template": plan_template["url"],
            "post_install_message_additional": plan_config.get(
                "post_install_message_additional", ""
            ),
            "preflight_message_additional": plan_config.get(
                "preflight_message_additional", ""
            ),
            "steps": steps,
            "supported_orgs": supported_orgs,
            "tier": plan_config["tier"],
            "title": plan_config["title"],
            "version": version["url"],
            # Use same AllowedList as the product, if any
            "visible_to": product.get("visible_to"),
        }
        if plan_config.get("checks"):
            plan_json["preflight_checks"] = plan_config["checks"]

        # Create Plan
        plan = self.api.create_plan(plan=plan_json)
        self.logger.info(f"Created Plan {plan['url']}")

    def _get_allowed_org_providers(self, plan_name: str) -> List[str]:
        "Validates and returns the org providers for a given plan"
        plan = Plan.parse_obj(self.project_config.config["plans"][plan_name])
        return plan.allowed_org_providers

    def _convert_org_providers_to_plan_equivalent(
        self, providers: Union[None, List[str]]
    ) -> str:
        """Given a list of plan.allowed_org_providers return the value that
        corresponds `supported_orgs` field on the Plan model in MetaDeploy
        """
        if not providers or providers == ["user"]:
            org_providers = "Persistent"
        elif "user" in providers and "devhub" in providers:
            org_providers = "Both"
        elif providers == ["devhub"]:
            org_providers = "Scratch"

        return org_providers

    def _find_product(self, repo_url):
        product = self.api.find_product(repo_url)
        self._add_labels(
            product,
            "product",
            {
                "title": "name of product",
                "short_description": "tagline of product",
                "description": "shown on product detail page (markdown)",
                "click_through_agreement": "legal text shown in modal dialog",
                "error_message": "shown after failed installation (markdown)",
            },
        )
        return product

    def _find_or_create_version(self, product):
        """Create a Version in MetaDeploy if it doesn't already exist"""
        if self.tag:
            label = self.project_config.get_version_for_tag(self.tag)
        else:
            label = self.commit

        version: Optional[dict] = self.api.find_version(
            query={"product": product["id"], "label": label}
        )
        verb: str = "Found" if version else "Created"

        if not version:
            version_dict = {
                "product": product["url"],
                "label": label,
                "description": self.options.get("description", ""),
                "is_production": True,
                "commit_ish": self.tag or self.commit,
                "is_listed": False,
            }
            version = self.api.create_version(version_dict)

        self.logger.info(f"{verb} {version['url']}")
        return version

    def _find_or_create_plan_template(self, product, plan_name, plan_config):
        plantemplate = self.api.find_plan_template(
            {"product": product["id"], "name": plan_name}
        )
        verb: str = "Found" if plantemplate else "Created"

        if not plantemplate:
            plantemplate = self.api.create_plan_template(
                {
                    "name": plan_name,
                    "product": product["url"],
                    "preflight_message": plan_config.get("preflight_message", ""),
                    "post_install_message": plan_config.get("post_install_message", ""),
                    "error_message": plan_config.get("error_message", ""),
                }
            )
            planslug = self.api.create_plan_slug(
                {"slug": plan_config["slug"], "parent": plantemplate["url"]}
            )
            self.logger.info(f"Created {planslug['url']}")
        self.logger.info(f"{verb} {plantemplate['url']}")
        return plantemplate

    def _load_labels(self):
        """Load existing English labels."""
        try:
            labels_path: Path = Path(self.labels_path, "labels_en.json")
            self.labels: dict = json.loads(labels_path.read_text())
        except FileNotFoundError:
            self.labels: dict = {}

    def _add_labels(self, obj, category, fields):
        """Add specified fields from obj to a label category."""
        for name, description in fields.items():
            if text := obj.get(name):
                if category not in self.labels:
                    self.labels[category] = {}
                label = {"message": text, "description": description}
                self.labels[category][name] = label

    def _add_label(self, category, text, description):
        """Add a single label to a label category."""
        if not text:
            return
        if category not in self.labels:
            self.labels[category] = {}
        label = {"message": text, "description": description}
        self.labels[category][text] = label

    def _save_labels(self):
        """Save updates to English labels."""
        with contextlib.suppress(FileNotFoundError):
            labels_path: Path = Path(self.labels_path, "labels_en.json")
            self.logger.info(f"Updating labels in {labels_path}")
            labels_path.write_text(json.dumps(self.labels, indent=4))

    def _publish_labels(self, slug: str):
        """Publish labels in all languages to MetaDeploy."""
        assert isinstance(slug, str), f"Slug should be a string, not {slug}"
        for path in Path(self.labels_path).glob("*.json"):
            lang = path.stem.split("_")[-1].lower()
            if lang in ("en", "en-us"):
                continue
            orig_labels = json.loads(path.read_text())
            prefixed_labels = {
                f"{slug}:{context}": labels for context, labels in orig_labels.items()
            }

            self.logger.info(f"Updating {lang} translations")
            try:
                self.api.update_lang_translation(lang, prefixed_labels)
            except requests.exceptions.HTTPError as err:
                self.logger.warning(f"Could not update {lang} translation: {err}")



import time
from contextlib import contextmanager
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryDirectory
from typing import Any
from unittest.mock import patch

import pytest
import yaml

from cumulusci.core.datasets import Dataset
from cumulusci.core.exceptions import BulkDataException
from cumulusci.salesforce_api.org_schema import Filters
from cumulusci.tasks.bulkdata.extract_dataset_utils.tests.test_synthesize_extract_declarations import (
    _fake_get_org_schema,
    describe_for,
)
from cumulusci.tasks.bulkdata.tests.integration_test_utils import (
    ensure_accounts,
    ensure_records,
)

ensure_accounts = ensure_accounts
ensure_records = ensure_records


@contextmanager
def setup_test(org_config):
    object_counts = {"Account": 6, "Contact": 1, "Opportunity": 5}
    obj_describes = (
        describe_for("Account"),
        describe_for("Contact"),
        describe_for("Opportunity"),
    )
    with patch.object(type(org_config), "is_person_accounts_enabled", False), patch(
        "cumulusci.core.datasets.get_org_schema",
        lambda _sf, org_config, **kwargs: _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            included_objects=["Account", "Contact", "Opportunity"],
            **kwargs,
        ),
    ):
        yield


class Timer:
    def __init__(self):
        self.old_time = time.time()

    def checkpoint(self, name):
        cur = time.time()
        print(name, cur - self.old_time)
        self.old_time = time.time()


@pytest.mark.vcr()
class TestDatasetsE2E:
    def test_datasets_e2e(
        self,
        sf,
        project_config,
        org_config,
        delete_data_from_org,
        ensure_accounts,
        run_code_without_recording,
    ):
        timer = Timer()
        timer.checkpoint("Started")
        object_counts = {"Account": 6, "Contact": 1, "Opportunity": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Opportunity"),
        )

        with patch.object(
            type(org_config), "is_person_accounts_enabled", False
        ), _fake_get_org_schema(
            org_config,
            obj_describes,
            object_counts,
            include_counts=True,
            filters=[Filters.extractable, Filters.createable],
            included_objects=["Account", "Contact", "Opportunity"],
        ) as schema, ensure_accounts(
            6
        ), Dataset(
            "foo", project_config, sf, org_config, schema=schema
        ) as dataset:
            timer.checkpoint("In Dataset")
            if dataset.path.exists():
                rmtree(dataset.path)

            self.demo_dataset(
                dataset, timer, sf, delete_data_from_org, run_code_without_recording
            )

            if dataset.path.exists():
                rmtree(dataset.path)

    def demo_dataset(
        self, dataset, timer, sf, delete_data_from_org, run_code_without_recording
    ):
        # Need record types for the RecordTypeId field to be in the org
        create_record_type_for_account(sf, run_code_without_recording)

        def count(sobject):
            return sf.query(f"select count(Id) from {sobject}")["records"][0]["expr0"]

        assert count("Account") == 6
        timer.checkpoint("Before create")
        # Create a dataset directory and associated files
        dataset.create()
        load_rules = [{"sf_object": "Account", "api": "rest"}]
        dataset.load_rules_file.write_text(yaml.safe_dump(load_rules))

        timer.checkpoint("After create")

        # Read selected schema objects and fields into a T.Dict[str, T.List[str]]
        # Datastructure like {"Account": ["Name", "Description"]}
        objs = dataset.read_schema_subset()
        timer.checkpoint("Read Subset")

        # Save a similar datastructure back to the file system
        objs = {"Account": objs["Account"]}
        objs["Account"].remove("Description")
        objs["Account"].remove("History__c")
        objs["Account"].remove("ns__Description__c")
        objs["Account"].remove("ns__LinkedAccount__c")
        objs["Account"].remove("Primary_Contact__c")
        dataset.update_schema_subset(objs)
        timer.checkpoint("Updated Subset")

        # Read full schema in a Datastructure like
        # {"Account": {"Name": True, "Description": False}}
        objs = dataset.read_which_fields_selected()
        assert objs["Account"]["Name"] is True
        assert objs["Account"]["Description"] is False
        timer.checkpoint("Read selected")

        # Run an extract to the filesystem
        with patch("cumulusci.tasks.bulkdata.extract.validate_and_inject_mapping"):
            dataset.extract()
        timer.checkpoint("Extract")

        run_code_without_recording(
            lambda: delete_data_from_org(
                [
                    "Account",
                    "Contact",
                ]
            )
        )
        timer.checkpoint("Deleted")

        assert count("Account") == 0
        dataset.load()
        timer.checkpoint("Loaded")

        assert count("Account") == 6
        timer.checkpoint("Verified")

    def test_datasets_extract_standard_objects(
        self,
        sf,
        project_config,
        org_config,
        delete_data_from_org,
        ensure_accounts,
        run_code_without_recording,
    ):
        timer = Timer()
        timer.checkpoint("Started")
        object_counts = {"Account": 6, "Contact": 1, "Opportunity": 5}
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Opportunity"),
        )
        # Need record types for the RecordTypeId field to be in the org
        create_record_type_for_account(sf, run_code_without_recording)

        with patch.object(type(org_config), "is_person_accounts_enabled", False), patch(
            "cumulusci.core.datasets.get_org_schema",
            lambda _sf, org_config, **kwargs: _fake_get_org_schema(
                org_config,
                obj_describes,
                object_counts,
                included_objects=["Account", "Contact", "Opportunity"],
                **kwargs,
            ),
        ), ensure_accounts(6), Dataset(
            "bar", project_config, sf, org_config
        ) as dataset:
            timer.checkpoint("In Dataset")
            if dataset.path.exists():
                rmtree(dataset.path)

            dataset.create()
            timer.checkpoint("After create")

            # Read selected schema objects and fields into a T.Dict[str, T.List[str]]
            # Datastructure like {"Account": ["Name", "Description"]}
            objs = dataset.read_schema_subset()
            timer.checkpoint("Read Subset")

            # Save a similar datastructure back to the file system
            objs = {
                objname: [f for f in fields if "__" not in f]
                for objname, fields in objs.items()
                if objname not in ["EdgeMartDataFile"] and "__" not in objname
            }
            dataset.update_schema_subset(objs)
            timer.checkpoint("Updated Subset")

            # Run an extract to the filesystem
            dataset.extract()
            timer.checkpoint("Extract")

            if dataset.path.exists():
                rmtree(dataset.path)

    def test_datasets_read_explicit_extract_declaration(
        self, sf, project_config, org_config, delete_data_from_org, ensure_accounts
    ):
        object_counts = {
            "Account": 6,
            "Contact": 1,
            "Opportunity": 5,
            "Lead": 1,
            "Event": 2,
        }
        obj_describes = (
            describe_for("Account"),
            describe_for("Contact"),
            describe_for("Opportunity"),
            describe_for("Lead"),
            describe_for("Event"),
        )
        with patch.object(type(org_config), "is_person_accounts_enabled", False), patch(
            "cumulusci.core.datasets.get_org_schema",
            lambda _sf, org_config, **kwargs: _fake_get_org_schema(
                org_config,
                obj_describes,
                object_counts,
                included_objects=["Account", "Contact", "Opportunity"],
                **kwargs,
            ),
        ), ensure_accounts(6), Dataset(
            "bar", project_config, sf, org_config
        ) as dataset:
            if dataset.path.exists():
                rmtree(dataset.path)

            dataset.create()
            with TemporaryDirectory() as t:

                def write_yaml(filename: str, json: Any):
                    full_path = Path(t) / filename
                    with open(full_path, "w") as f:
                        yaml.safe_dump(json, f)
                    return full_path

                extract_def = write_yaml(
                    "test_extract_definition",
                    {
                        "extract": {
                            "Account": {"fields": ["Name"]},
                            "Contact": {
                                "fields": ["FirstName", "LastName", "AccountId"]
                            },
                            "Event": {"fields": ["Subject", "WhoId"]},
                        }
                    },
                )
                loading_rules = write_yaml(
                    "loading_rules.load.yml",
                    [
                        {"sf_object": "Account", "load_after": "Contact"},
                        {"sf_object": "Lead", "load_after": "Event"},
                    ],
                )

                # Don't actually extract data.
                with patch("cumulusci.tasks.bulkdata.ExtractData._run_task"):
                    dataset.extract(
                        extraction_definition=extract_def,
                        loading_rules_file=loading_rules,
                    )
            with open(dataset.mapping_file) as p:
                actual = yaml.safe_load(p)
                expected = {
                    "Insert Contact": {
                        "sf_object": "Contact",
                        "table": "Contact",
                        "fields": ["FirstName", "LastName"],
                        "lookups": {
                            "AccountId": {
                                "table": ["Account"],
                                "key_field": "AccountId",
                                "after": "Insert Account",
                            }
                        },
                    },
                    "Insert Event": {
                        "sf_object": "Event",
                        "table": "Event",
                        "fields": ["Subject"],
                        "lookups": {
                            "WhoId": {
                                "table": ["Contact", "Lead"],
                                "key_field": "WhoId",
                                "after": "Insert Lead",
                            }
                        },
                    },
                    "Insert Account": {
                        "sf_object": "Account",
                        "table": "Account",
                        "fields": ["Name"],
                    },
                    "Insert Lead": {
                        "sf_object": "Lead",
                        "table": "Lead",
                        "fields": ["Company", "LastName"],
                    },
                }
                assert tuple(actual.items()) == tuple(expected.items()), actual.items()

            if dataset.path.exists():
                rmtree(dataset.path)


class TestLoadDatasets:
    def test_dataset_with_snowfakery(self, sf, project_config, org_config):
        def fake_path_exists(self):
            return str(self).endswith("recipe.yml")

        called = False

        def fake_run_snowfakery(self):
            nonlocal called
            assert "foo.recipe.yml" in self.options["recipe"]
            called = True

        with setup_test(org_config), Dataset(
            "foo", project_config, sf, org_config, schema=None
        ) as dataset, patch(
            "cumulusci.core.datasets.Path.exists", fake_path_exists
        ), patch(
            "cumulusci.tasks.bulkdata.snowfakery.Snowfakery._run_task",
            fake_run_snowfakery,
        ):
            dataset.load()
        assert called

    def test_dataset_with_no_data_or_recipe(self, sf, project_config, org_config):
        with setup_test(org_config), Dataset(
            "fxoyoxz", project_config, sf, org_config, schema=None
        ) as dataset, pytest.raises(BulkDataException, match="fxoyoxz"):
            dataset.load()


# Need record types for the RecordTypeId field to be in the org
def create_record_type_for_account(sf, run_code_without_recording):
    def create_record_type_for_account_real():
        account_record_types = sf.query(
            "select Count(Id) from RecordType where SObjectType='Account'"
        )
        if account_record_types["records"][0]["expr0"] == 0:
            sf.RecordType.create(
                {
                    "DeveloperName": "PytestAccountRecordType",
                    "Name": "PytestAccountRecordType",
                    "SObjectType": "Account",
                }
            )

    run_code_without_recording(create_record_type_for_account_real)



from datetime import datetime
from unittest import mock

import pytest

from cumulusci.core.exceptions import CumulusCIException, TaskOptionsError
from cumulusci.tasks.metadata_etl import SetOrgWideDefaults
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.utils.xml import metadata_tree

MD = "{%s}" % metadata_tree.METADATA_NAMESPACE

CUSTOMOBJECT_XML = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <sharingModel>Read</sharingModel>
    <externalSharingModel>Read</externalSharingModel>
    <label>Test</label>
    <pluralLabel>Tests</pluralLabel>
    <nameField>
        <label>Test Name</label>
        <trackHistory>false</trackHistory>
        <type>Text</type>
    </nameField>
    <deploymentStatus>Deployed</deploymentStatus>
</CustomObject>"""

CUSTOMOBJECT_XML_MISSING_TAGS = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <label>Test</label>
    <pluralLabel>Tests</pluralLabel>
    <nameField>
        <label>Test Name</label>
        <trackHistory>false</trackHistory>
        <type>Text</type>
    </nameField>
    <deploymentStatus>Deployed</deploymentStatus>
</CustomObject>"""


class TestSetOrgWideDefaults:
    def test_sets_owd(self):
        task = create_task(
            SetOrgWideDefaults,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "org_wide_defaults": [
                    {
                        "api_name": "Account",
                        "internal_sharing_model": "Private",
                        "external_sharing_model": "Private",
                    },
                    {
                        "api_name": "Test__c",
                        "internal_sharing_model": "ReadWrite",
                        "external_sharing_model": "Read",
                    },
                ],
            },
        )

        assert task.api_names == set(["Account", "Test__c"])

        tree = metadata_tree.fromstring(CUSTOMOBJECT_XML)

        result = task._transform_entity(tree, "Test__c")

        entry = result._element.findall(f".//{MD}sharingModel")
        assert len(entry) == 1
        assert entry[0].text == "ReadWrite"

        entry = result._element.findall(f".//{MD}externalSharingModel")
        assert len(entry) == 1
        assert entry[0].text == "Read"

    def test_sets_owd__missing_tags(self):
        task = create_task(
            SetOrgWideDefaults,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "org_wide_defaults": [
                    {
                        "api_name": "Account",
                        "internal_sharing_model": "Private",
                        "external_sharing_model": "Private",
                    },
                    {
                        "api_name": "Test__c",
                        "internal_sharing_model": "ReadWrite",
                        "external_sharing_model": "Read",
                    },
                ],
            },
        )

        assert task.api_names == set(["Account", "Test__c"])

        tree = metadata_tree.fromstring(CUSTOMOBJECT_XML_MISSING_TAGS)

        result = task._transform_entity(tree, "Test__c")._element

        entry = result.findall(f".//{MD}sharingModel")
        assert len(entry) == 1
        assert entry[0].text == "ReadWrite"

        entry = result.findall(f".//{MD}externalSharingModel")
        assert len(entry) == 1
        assert entry[0].text == "Read"

    def test_post_deploy_waits_for_enablement(self):
        task = create_task(
            SetOrgWideDefaults,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "org_wide_defaults": [
                    {
                        "api_name": "Account",
                        "internal_sharing_model": "Private",
                        "external_sharing_model": "Private",
                    },
                    {
                        "api_name": "Test__c",
                        "internal_sharing_model": "ReadWrite",
                        "external_sharing_model": "Read",
                    },
                ],
            },
        )
        task.sf = mock.Mock()
        task.sf.query.side_effect = [
            {
                "totalSize": 1,
                "records": [
                    {"ExternalSharingModel": "Read", "InternalSharingModel": "Read"}
                ],
            },
            {
                "totalSize": 1,
                "records": [
                    {
                        "ExternalSharingModel": "Private",
                        "InternalSharingModel": "Private",
                    }
                ],
            },
            {
                "totalSize": 1,
                "records": [
                    {
                        "ExternalSharingModel": "Read",
                        "InternalSharingModel": "ReadWrite",
                    }
                ],
            },
        ]
        task._post_deploy("Success")

        query = (
            "SELECT ExternalSharingModel, InternalSharingModel "
            "FROM EntityDefinition "
            "WHERE QualifiedApiName = '{}'"
        )

        task.sf.query.assert_has_calls(
            [
                mock.call(query.format("Account")),
                mock.call(query.format("Account")),
                mock.call(query.format("Test__c")),
            ]
        )

        assert task.poll_complete

    def test_post_deploy_waits_for_enablement__namespaced_org(self):
        task = create_task(
            SetOrgWideDefaults,
            {
                "managed": False,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "org_wide_defaults": [
                    {
                        "api_name": "Account",
                        "internal_sharing_model": "Private",
                        "external_sharing_model": "Private",
                    },
                    {
                        "api_name": "Test__c",
                        "internal_sharing_model": "ReadWrite",
                        "external_sharing_model": "Read",
                    },
                ],
            },
        )
        task.org_config.namespaced = True
        task.project_config.project__package__namespace = "test"
        task.sf = mock.Mock()
        task.sf.query.side_effect = [
            {
                "totalSize": 1,
                "records": [
                    {"ExternalSharingModel": "Read", "InternalSharingModel": "Read"}
                ],
            },
            {
                "totalSize": 1,
                "records": [
                    {
                        "ExternalSharingModel": "Private",
                        "InternalSharingModel": "Private",
                    }
                ],
            },
            {
                "totalSize": 1,
                "records": [
                    {
                        "ExternalSharingModel": "Read",
                        "InternalSharingModel": "ReadWrite",
                    }
                ],
            },
        ]
        task._post_deploy("Success")

        query = (
            "SELECT ExternalSharingModel, InternalSharingModel "
            "FROM EntityDefinition "
            "WHERE QualifiedApiName = '{}'"
        )

        task.sf.query.assert_has_calls(
            [
                mock.call(query.format("Account")),
                mock.call(query.format("Account")),
                mock.call(query.format("test__Test__c")),
            ]
        )

        assert task.poll_complete

    def test_post_deploy_exception_not_found(self):
        task = create_task(
            SetOrgWideDefaults,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "org_wide_defaults": [
                    {
                        "api_name": "Account",
                        "internal_sharing_model": "Private",
                        "external_sharing_model": "Private",
                    }
                ],
            },
        )
        task.sf = mock.Mock()
        task.sf.query.return_value = {"totalSize": 0, "records": []}
        with pytest.raises(CumulusCIException):
            task._post_deploy("Success")

        query = (
            "SELECT ExternalSharingModel, InternalSharingModel "
            "FROM EntityDefinition "
            "WHERE QualifiedApiName = '{}'"
        )

        task.sf.query.assert_has_calls([mock.call(query.format("Account"))])

    def test_raises_exception_timeout(self):
        task = create_task(
            SetOrgWideDefaults,
            {
                "managed": True,
                "api_version": "47.0",
                "api_names": "bar,foo",
                "org_wide_defaults": [
                    {
                        "api_name": "Account",
                        "internal_sharing_model": "Private",
                        "external_sharing_model": "Private",
                    }
                ],
            },
        )

        task.time_start = datetime.min
        with pytest.raises(CumulusCIException):
            task._poll_action()

    def test_raises_exception_missing_values(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                SetOrgWideDefaults,
                {
                    "managed": True,
                    "api_version": "47.0",
                    "api_names": "bar,foo",
                    "org_wide_defaults": [
                        {
                            "api_name": "Account",
                            "internal_sharing_model": "Private",
                            "external_sharing_model": "Private",
                        },
                        {"api_name": "Test__c"},
                    ],
                },
            )

    def test_raises_exception_bad_sharing_model(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                SetOrgWideDefaults,
                {
                    "managed": True,
                    "api_version": "47.0",
                    "api_names": "bar,foo",
                    "org_wide_defaults": [
                        {
                            "api_name": "Account",
                            "internal_sharing_model": "Nonsense",
                            "external_sharing_model": "Private",
                        },
                        {"api_name": "Test__c"},
                    ],
                },
            )



import contextlib
import io
import json
from unittest.mock import Mock, patch

import click
import pytest

from cumulusci.cli.runtime import CliRuntime
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.exceptions import CumulusCIUsageError

from .. import task
from .utils import DummyTask, run_click_command

color_opts = {"options": {"color": {}}}
multiple_opts = {"options": {"foo": {}, "bar": {}, "baz": {}}}

test_tasks = {
    "dummy-task": {
        "class_path": "cumulusci.cli.tests.utils.DummyTask",
        "description": "This is a dummy task.",
    },
    "dummy-derived-task": {
        "class_path": "cumulusci.cli.tests.test_run_task.DummyDerivedTask"
    },
}


@pytest.fixture
def runtime():
    runtime = CliRuntime(load_keychain=False)
    runtime.project_config.config["tasks"] = {**test_tasks}

    runtime.keychain = Mock()
    runtime.keychain.get_default_org.return_value = (None, None)
    runtime._load_keychain = Mock()

    yield runtime


def test_task_run(runtime):
    DummyTask._run_task = Mock()
    multi_cmd = task.RunTaskCommand()
    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "dummy-task")
        cmd.callback(runtime, "dummy-task", color="blue")

    DummyTask._run_task.assert_called_once()


def test_task_run__no_project(runtime):
    runtime.project_config = None
    runtime.project_config_error = Exception("Broken")
    multi_cmd = task.RunTaskCommand()
    with pytest.raises(Exception, match="Broken"):
        with click.Context(multi_cmd, obj=runtime) as ctx:
            multi_cmd.get_command(ctx, "dummy-task")


def test_task_run__debug_before(runtime):
    DummyTask._run_task = Mock()
    multi_cmd = task.RunTaskCommand()
    set_trace = Mock(side_effect=SetTrace)

    with click.Context(multi_cmd, obj=runtime) as ctx:
        with patch("pdb.set_trace", set_trace):
            with pytest.raises(SetTrace):
                cmd = multi_cmd.get_command(ctx, "dummy-task")
                cmd.callback(
                    runtime,
                    "dummy_task",
                    color="blue",
                    debug_before=True,
                    debug_after=False,
                )


def test_task_run__debug_after(runtime):
    DummyTask._run_task = Mock()
    multi_cmd = task.RunTaskCommand()
    set_trace = Mock(side_effect=SetTrace)

    with click.Context(multi_cmd, obj=runtime) as ctx:
        with patch("pdb.set_trace", set_trace):
            with pytest.raises(SetTrace):
                cmd = multi_cmd.get_command(ctx, "dummy-task")
                cmd.callback(
                    runtime,
                    "dummy-task",
                    color="blue",
                    debug_before=False,
                    debug_after=True,
                )


def test_task_run__help(runtime):
    DummyTask._run_task = Mock()
    multi_cmd = task.RunTaskCommand()
    with click.Context(multi_cmd, obj=runtime) as ctx:
        cmd = multi_cmd.get_command(ctx, "dummy-task")

    assert "This is a dummy task." in cmd.help  # task description


def test_task_run__list_commands(runtime):
    multi_cmd = task.RunTaskCommand()
    with click.Context(multi_cmd, obj=runtime) as ctx:
        commands = multi_cmd.list_commands(ctx)
    assert commands == ["dummy-derived-task", "dummy-task"]


def test_format_help(runtime):
    runtime.universal_config = Mock()
    multi_cmd = task.RunTaskCommand()
    with contextlib.redirect_stdout(io.StringIO()) as stdout:
        with click.Context(multi_cmd, obj=runtime) as ctx:
            multi_cmd.format_help(ctx, Mock())

    stdout = stdout.getvalue()
    assert "Usage: cci task run <task_name> [TASK_OPTIONS...]" in stdout
    assert "See above for a complete list of available tasks." in stdout
    assert "Use cci task info <task_name>" in stdout
    assert len(runtime.universal_config.method_calls) == 0


def test_get_default_command_options():
    opts = task.RunTaskCommand()._get_default_command_options(is_salesforce_task=False)
    assert len(opts) == 4

    opts = task.RunTaskCommand()._get_default_command_options(is_salesforce_task=True)
    assert len(opts) == 5
    assert any([o.name == "org" for o in opts])


def test_collect_task_options():
    new_options = {"debug-before": None}
    old_options = (("color", "green"),)

    opts = task.RunTaskCommand()._collect_task_options(
        new_options, old_options, "dummy-task", color_opts["options"]
    )
    assert opts == {"color": "green"}


def test_collect_task_options__duplicate():
    new_options = {"color": "aqua"}
    old_options = (("color", "green"),)

    with pytest.raises(CumulusCIUsageError):
        task.RunTaskCommand()._collect_task_options(
            new_options, old_options, "dummy-task", color_opts["options"]
        )


def test_collect_task_options__not_in_task():
    new_options = {}
    old_options = (("color", "green"),)

    with pytest.raises(CumulusCIUsageError):
        task.RunTaskCommand()._collect_task_options(
            new_options, old_options, "dummy-task", {"not-color": {}}
        )


@patch("cumulusci.cli.task.CliTable")
def test_task_list(cli_tbl):
    runtime = Mock()
    runtime.universal_config.cli__plain_output = None
    runtime.get_available_tasks.return_value = [
        {"name": "test_task", "description": "Test Task", "group": "Test Group"}
    ]

    run_click_command(task.task_list, runtime=runtime, plain=False, print_json=False)

    cli_tbl.assert_called_with(
        [["Task", "Description"], ["test_task", "Test Task"]], "Test Group"
    )


def test_task_list__json(capsys):
    expected_output = {
        "name": "test_task",
        "description": "This can be a really long description that might need a newline if some library is formatting it for ANSI output.",
        "group": "Test Group",
    }
    runtime = Mock()
    runtime.universal_config.cli__plain_output = None
    runtime.get_available_tasks.return_value = [expected_output]

    run_click_command(task.task_list, runtime=runtime, plain=False, print_json=True)

    captured = capsys.readouterr()
    parsed_output = json.loads(captured.out)[0]
    assert parsed_output == expected_output


@patch("cumulusci.cli.task.doc_task", return_value="docs")
def test_task_doc(doc_task):
    runtime = Mock()
    runtime.universal_config.tasks = {"test": {}}
    run_click_command(task.task_doc, runtime=runtime, project=False)
    doc_task.assert_called()


def test_task_doc__project__outside_project():
    runtime = Mock()
    runtime.project_config = None
    with pytest.raises(click.UsageError):
        run_click_command(task.task_doc, runtime=runtime, project=True)


@patch("click.echo")
@patch("cumulusci.cli.task.doc_task", return_value="docs")
def test_task_doc_project(doc_task, echo):
    runtime = Mock()
    runtime.universal_config = {"tasks": {}}
    runtime.project_config = BaseProjectConfig(
        runtime.universal_config,
        {
            "project": {"name": "Test"},
            "tasks": {"task1": {"a": "b"}, "task2": {}},
        },
    )
    runtime.project_config.config_project = {"tasks": {"task1": {"a": "b"}}}
    run_click_command(task.task_doc, runtime=runtime, project=True)
    doc_task.assert_called()
    echo.assert_called()


@patch("cumulusci.cli.task.Path")
@patch("click.echo")
@patch("cumulusci.cli.task.doc_task", return_value="docs")
def test_task_doc_project_write(doc_task, echo, Path):
    runtime = Mock()
    runtime.universal_config.tasks = {"test": {}}
    runtime.project_config = BaseProjectConfig(
        runtime.universal_config,
        {
            "project": {"name": "Test"},
            "tasks": {"option": {"a": "b"}},
        },
    )
    runtime.project_config.config_project = {"tasks": {"option": {"a": "b"}}}
    run_click_command(task.task_doc, runtime=runtime, project=True, write=True)
    doc_task.assert_called()
    echo.assert_not_called()


@patch("cumulusci.cli.task.rst2ansi")
@patch("cumulusci.cli.task.doc_task")
def test_task_info(doc_task, rst2ansi):
    runtime = Mock()
    runtime.project_config.tasks__test = {"options": {}}
    run_click_command(task.task_info, runtime=runtime, task_name="test")
    doc_task.assert_called_once()
    rst2ansi.assert_called_once()


class SetTrace(Exception):
    pass


class DummyDerivedTask(DummyTask):
    def _run_task(self):
        click.echo(f"<{self.__class__}>\n\tcolor: {self.options['color']}")



import base64
import hashlib
import logging
import os
from pathlib import Path
from typing import Optional, Tuple

from github3.git import Commit, Tree

from cumulusci.core.exceptions import GithubException


class CommitDir(object):
    """Commit all changes in local_dir to branch/repo_dir"""

    def __init__(self, repo, logger=None, author=None):
        """
        repo: github3.py Repository object, authenticated
        logger (optional): python logger object
        author (optional): dict of name, email, time - default uses auth'd info
        """
        self.repo = repo
        if not logger:
            logger = logging.getLogger(os.path.basename(__file__))
        self.logger = logger
        self.author = author or {}

    def _validate_dirs(self, local_dir, repo_dir):
        local_dir = os.path.abspath(local_dir)
        if not os.path.isdir(local_dir):
            raise GithubException(f"Not a dir: {local_dir}")
        # do not use os.path because repo_dir is not local
        if repo_dir is None:
            repo_dir = ""
        if repo_dir.startswith("."):
            repo_dir = repo_dir[1:]
        if repo_dir.startswith("/"):
            repo_dir = repo_dir[1:]
        if repo_dir.endswith("/"):
            repo_dir = repo_dir[:-1]
        return local_dir, repo_dir

    def __call__(
        self, local_dir, branch, repo_dir=None, commit_message=None, dry_run=False
    ):
        """
        local_dir: path to local directory to commit
        branch: target branch name
        repo_dir: target path within repo - use '' for repo root
        commit_message: message for git commit
        dry_run: skip creating GitHub data if True
        """

        self.dry_run = dry_run
        # prepare dir args
        self.local_dir, self.repo_dir = self._validate_dirs(local_dir, repo_dir)
        self._set_git_data(branch)

        self.new_tree_list = [self._create_new_tree_item(item) for item in self.tree]
        self.new_tree_list = [item for item in self.new_tree_list if item]
        self._add_new_files_to_tree(self.new_tree_list)

        if not self.new_tree_list:
            self.logger.warning("No changes found, aborting commit")
            return self.parent_commit

        self._summarize_changes(self.tree, self.new_tree_list)
        new_tree = self._create_tree(self.new_tree_list)
        new_commit = self._create_commit(commit_message, new_tree)
        self._update_head(new_commit)
        return new_commit

    def _set_git_data(self, branch):
        # get ref to branch HEAD
        self.head = self.repo.ref(f"heads/{branch}")

        # get commit pointed to by HEAD
        self.parent_commit: Commit = self.repo.git_commit(self.head.object.sha)

        # get tree of commit as dict
        orig_tree: Tree = self.repo.tree(self.parent_commit.tree.sha, recursive=True)
        self.tree = [
            git_hash.as_dict() for git_hash in orig_tree.tree if git_hash.type != "tree"
        ]

    def _create_new_tree_item(self, item: dict) -> Optional[dict]:
        """Given an Git tree element, returns None if it is unchanged, or
        a new element if it has been deleted or updated.
        """
        if not item["path"].startswith(self.repo_dir):
            # outside target dir in repo - keep in tree
            self.logger.debug(f'Unchanged (outside target path): {item["path"]}')
            return None

        local_path, content = self._find_and_read_item(item)
        new_item = item.copy()
        if content is None:
            # delete blob from tree by setting 'sha' to null
            self.logger.debug(f"Delete: {item['path']}")
            new_item["sha"] = None
        elif self._item_changed(item, content):
            # we need to delete the sha key because GH returns an error if
            # both 'sha' and 'contents' are passed.
            self.logger.debug(f"Update: {local_path}")
            new_item.pop("sha", None)
            new_item.update(self._get_content_or_sha(content, self.dry_run))
        else:
            self.logger.debug(f'Unchanged: {item["path"]}')
            return None
        return new_item

    def _add_new_files_to_tree(self, new_tree_list):
        new_tree_target_subpaths = [
            self._get_item_sub_path(item)
            for item in self.tree
            if item["path"].startswith(self.repo_dir)
        ]

        for root, _, files in os.walk(self.local_dir):
            for filename in files:
                if not filename.startswith("."):
                    local_file = os.path.join(root, filename)
                    local_file_subpath = local_file[(len(self.local_dir) + 1) :]
                    if local_file_subpath not in new_tree_target_subpaths:
                        repo_path = f"{self.repo_dir}/" if self.repo_dir else ""
                        new_item = {
                            "path": f'{repo_path}{local_file_subpath.replace(os.sep, "/")}',
                            "mode": "100644",  # FIXME: This is wrong
                            "type": "blob",
                        }
                        new_item.update(
                            self._get_content_or_sha(
                                Path(local_file).read_bytes(), self.dry_run
                            )
                        )
                        new_tree_list.append(new_item)

    def _summarize_changes(self, old_tree_list, new_tree_list):
        self.logger.info("Summary of changes:")
        old_paths = [item["path"] for item in old_tree_list]

        def join_paths(paths):
            return "\n".join(f"\t{path}" for path in paths)

        deleted_paths = join_paths(
            [item["path"] for item in new_tree_list if item.get("sha", "") is None]
        )
        if deleted_paths:
            self.logger.warning(f"Deleted:\n{deleted_paths}")

        new_paths = join_paths(
            [
                item["path"]
                for item in new_tree_list
                if "content" in item and item["path"] not in old_paths
            ]
        )
        if new_paths:
            self.logger.info(f"Added:\n{new_paths}")

        updated_paths = join_paths(
            [
                item["path"]
                for item in new_tree_list
                if "content" in item and item["path"] in old_paths
            ]
        )
        if updated_paths:
            self.logger.info(f"Updated:\n{updated_paths}")

    def _create_tree(self, new_tree_list):
        new_tree = None
        if self.dry_run:
            self.logger.info("[dry_run] Skipping creation of new tree")
        else:
            self.logger.info("Creating new tree")
            new_tree = self.repo.create_tree(
                new_tree_list, base_tree=self.parent_commit.tree.sha
            )
            if not new_tree:
                raise GithubException("Failed to create tree")
        return new_tree

    def _create_commit(self, commit_message, new_tree):
        if commit_message is None:
            commit_message = f"Commit dir {self.local_dir} to {self.repo.owner}/{self.repo.name}/{self.repo_dir} via CumulusCI"

        new_commit = None
        if self.dry_run:
            self.logger.info("[dry_run] Skipping creation of new commit")
        else:
            self.logger.info("Creating new commit")
            commit_info = {
                "message": commit_message,
                "tree": new_tree.sha,
                "parents": [self.parent_commit.sha],
            }
            if self.author:
                commit_info["author"] = self.author
                commit_info["committer"] = self.author
            new_commit = self.repo.create_commit(**commit_info)
            if not new_commit:
                raise GithubException("Failed to create commit")
        return new_commit

    def _update_head(self, new_commit):
        if self.dry_run:
            self.logger.info("[dry_run] Skipping call to update HEAD")
        else:
            self.logger.info("Updating HEAD")
            success = self.head.update(new_commit.sha)
            if not success:
                raise GithubException("Failed to update HEAD")

    def _get_item_sub_path(self, item):
        len_path = (len(self.repo_dir) + 1) if self.repo_dir else 0
        return item["path"][len_path:]

    def _find_and_read_item(self, item) -> Tuple[Path, Optional[bytes]]:
        item_subpath = self._get_item_sub_path(item)
        pth = Path(self.local_dir, item_subpath)
        contents = pth.read_bytes() if pth.exists() else None
        return pth, contents

    def _item_changed(self, item: dict, content: bytes) -> bool:
        header = b"blob " + str(len(content)).encode() + b"\0"
        return hashlib.sha1(header + content).hexdigest() != item["sha"]

    def _get_content_or_sha(self, content, dry_run) -> dict:
        """Given file contents as bytes, returns a dict with file contents.

        If the contents are text: {'contents': contents as str}
        If the contents are binary: {'sha': hash of the uploaded blob}
        """
        if dry_run:
            self.logger.info("[dry_run] Skipping creation of blob for new file")
            return {"content": None}
        try:
            content = content.decode("utf-8")
            return {"content": content}
        except UnicodeDecodeError:
            return self._create_blob(content)

    def _create_blob(self, contents: bytes) -> dict:
        self.logger.info("Creating blob for binary file")
        content: bytes = base64.b64encode(contents)
        blob_sha: str = self.repo.create_blob(content.decode("utf-8"), "base64")
        if not blob_sha:
            raise GithubException("Failed to create blob")
        self.logger.debug(f"Blob created: {blob_sha}")
        return {"sha": blob_sha}
